/* thttpd.c - tiny/turbo/throttling HTTP server
**
** Copyright  1995,1998,1999,2000,2001 by Jef Poskanzer <jef@acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

#include "config.h"
#include "version.h"

#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include <errno.h>
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <pwd.h>
#ifdef HAVE_GRP_H
#include <grp.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#ifdef TIME_WITH_SYS_TIME
#include <time.h>
#endif
#include <unistd.h>

#include "fdwatch.h"
#include "libhttpd.h"
#include "mmc.h"
//#include "timers.h"
#include "match.h"

#ifdef STATISTICS
#include <userManager.h>
#endif
#ifdef NAS
#include <NAS.h>
#endif
#ifdef DVR
#include <Util.h>
#endif

static char* argv0;
static int debug;
static int port;
static char* dir;
static int do_chroot, no_log, no_symlink, do_vhost, do_global_passwd;
static char* cgi_pattern;
static char* url_pattern;
static int no_empty_referers;
static char* local_pattern;
static char* logfile;
static char* throttlefile;
static char* hostname;
static char* pidfile;
static char* user;
static char* charset;
static int   OccasionalMmcTime = OCCASIONAL_MMC_TIME;
static long  MlsIdleSendRespTimeLimit  = 1000L * IDLE_SEND_RESP_TIMELIMIT;
static long  MlsIdleKeepAliveTimeLimit = 1000L * IDLE_KEEPALIVE_TIMELIMIT;

typedef struct {
    char* pattern;
    long limit;
    long rate;
    off_t bytes_since_avg;
    int num_sending;
    } throttletab;
static throttletab* throttles;
static int numthrottles, maxthrottles;

#define THROTTLE_NOLIMIT 1234567890L	/* default limit (if none) */

#ifdef USE_LAYOUT
/* Layout Vars, global for now */
char *lheaderfile, *lfooterfile;
int   lheaderfile_len, lfooterfile_len;
char *lheaderfile_map, *lfooterfile_map;
#endif /* USE_LAYOUT */

typedef struct {
    int conn_state;
    httpd_conn* hc;
    int tnums[MAXTHROTTLENUMS];         /* throttle indexes */
    int numtnums;
    int keep_alive;
    long limit;
    time_t started_at;
    Timer* idle_read_timer;
    Timer* idle_send_timer;
    Timer* wakeup_timer;
    Timer* linger_timer;
    long wouldblock_delay;
    off_t bytes;
    off_t bytes_to_send;
    off_t bytes_sent;
    } connecttab;
static connecttab* connects;
static int numconnects, maxconnects;
static int httpd_conn_count;

/* The connection states. */
#define CNST_FREE		0
#define CNST_READING		1
#define CNST_SENDING		2
#define CNST_SENDING_RESP	3
#define CNST_PAUSING		4
#define CNST_LINGERING		5

/* boolean values for do keep alive in clear_connection() */
#define NO_KEEP_ALIVE	0	/* FALSE */
#define DO_KEEP_ALIVE	1	/* TRUE */

#ifdef	HAVE_SIGSET
#define my_signal	sigset
#else
#define my_signal	signal
#endif

/* flags for gotSigMask */

#define GOT_O_SIGHUP	0x0001
#define GOT_O_SIGUSR1	0x0002
#define GOT_O_SIGUSR2	0x0004

static httpd_server* hs = (httpd_server*) 0;
static int terminate;
static time_t start_time, stats_time;
static long stats_connections, stats_bytes;
static int stats_simultaneous;

static volatile sig_atomic_t in_shut_down;
static volatile sig_atomic_t gotSigMask;


/* Forwards. */
static void parse_args( int argc, char** argv );
static void usage( void );
#ifdef USE_LAYOUT
	static void* map_layoutfile( char* filename, int* pfilesize );
#endif
static void read_config( char* filename );
static void value_required( char* name, char* value );
static void no_value_required( char* name, char* value );
static char* e_strdup( char* oldstr );
static void lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockaddr* sa6P, size_t sa6_len, int* gotv6P );
static void read_throttlefile( char* throttlefile );
static void shut_down( void );
static int handle_newconnect( struct timeval* tvP, int listen_fd );
static void handle_buf_read( connecttab* c, struct timeval* tvP );
static void handle_read( connecttab* c, struct timeval* tvP );
static void handle_send( connecttab* c, struct timeval* tvP );
static void handle_send_resp( connecttab* c, struct timeval* tvP );
static void handle_linger( connecttab* c, struct timeval* tvP );
static int check_throttles( connecttab* c );
static void clear_throttles( connecttab* c, struct timeval* tvP );
static void update_throttles( ClientData client_data, struct timeval* nowP );
static void resp_clear_connection( connecttab* c, struct timeval* tvP );
static void clear_connection( connecttab* c, struct timeval* tvP, int doKeep );
static void really_clear_connection( connecttab* c, struct timeval* tvP );
static void idle_read_connection( ClientData client_data, struct timeval* nowP );
static void idle_send_connection( ClientData client_data, struct timeval* nowP );
static void idle_send_resp_connection( ClientData client_data, struct timeval* nowP );
static void wakeup_connection( ClientData client_data, struct timeval* nowP );
static void linger_clear_connection( ClientData client_data, struct timeval* nowP );
static void occasional_mmc( ClientData client_data, struct timeval* nowP );
static void occasional_tmr( ClientData client_data, struct timeval* nowP );
#ifdef STATS_TIME
static void show_stats( ClientData client_data, struct timeval* nowP );
#endif /* STATS_TIME */
static void logstats( struct timeval* nowP );
static void thttpd_logstats( long secs );

static void
thttpd_log_reopen( void )
    {
    FILE* logfp;

    if ( no_log )
	return;

    /* Re-open the log file. */
    if ( logfile != (char*) 0 )
	{
	logfp = fopen( logfile, "a" );
	if ( logfp == (FILE*) 0 )
	    {
	    syslog( LOG_CRIT, "reopening %.80s - %m", logfile );
	    return;
	    }
	(void) fcntl( fileno( logfp ), F_SETFD, 1 );
	httpd_set_logfp( hs, logfp );
	}
    }


static void
handle_rsig( int sig )
    {
    syslog( LOG_WARNING, "handle_rsig: signal %d, shutdown interrupted", sig );
    closelog();
    exit( sig );
    }

static void
handle_term( int sig )
    {
    syslog( LOG_NOTICE, "handle_term: signal %d, shutting down ...", sig );
    /* if the same signal occurs in shut_down() then exit directly */
    /* in order to avoid recursive calls to this function */
    my_signal( sig, handle_rsig );
    if ( in_shut_down )
	handle_rsig( sig );
    shut_down();
    syslog( LOG_NOTICE, "handle_term: shutdown completed, exit( %d )", sig );
    closelog();
    exit( sig );
    }


static void
handle_hup( int sig )
    {
    gotSigMask |= GOT_O_SIGHUP;
#ifndef	HAVE_SIGSET
    my_signal( sig, handle_hup );
#endif
    }


static void
handle_usr1( int sig )
    {
    gotSigMask |= GOT_O_SIGUSR1;
#ifndef	HAVE_SIGSET
    my_signal( sig, handle_usr1 );
#endif
    }


static void
handle_usr2( int sig )
    {
    gotSigMask |= GOT_O_SIGUSR2;
#ifndef	HAVE_SIGSET
    my_signal( sig, handle_usr2 );
#endif
    }


int __block_request=0;

static void
handle_block_req( int sig )
{
    __block_request=1;
}

int
main( int argc, char** argv )
    {
    char* cp;
    struct passwd* pwd;
    uid_t uid;
    gid_t gid;
    char cwd[MAXPATHLEN];
    FILE* logfp;
    int num_ready;
    int cnum, ridx;
    connecttab* c;
    httpd_conn* hc;
    httpd_sockaddr sa4;
    httpd_sockaddr sa6;
    int gotv4 = 0, gotv6 = 0;
    struct timeval tv = { 0, 0 };

    argv0 = argv[0];

    cp = strrchr( argv0, '/' );
    if ( cp != (char*) 0 )
	++cp;
    else
	cp = argv0;
    openlog( cp, LOG_NDELAY|LOG_PID, LOG_FACILITY );

    /* Handle command-line arguments. */
    parse_args( argc, argv );

    /* Check port number. */
    if ( port <= 0 )
	{
	syslog( LOG_CRIT, "illegal port number" );
	(void) fprintf( stderr, "%s: illegal port number\n", argv0 );
	exit( 40 );
	}

    /* Read zone info now, in case we chroot(). */
    tzset();

    /* Look up hostname now, in case we chroot(). */
    lookup_hostname( &sa4, sizeof(sa4), &gotv4, &sa6, sizeof(sa6), &gotv6 );
    if ( ! ( gotv4 || gotv6 ) )
	{
	syslog( LOG_ERR, "can't find any valid address" );
	(void) fprintf( stderr, "%s: can't find any valid address\n", argv0 );
	exit( 41 );
	}

    /* Throttle file. */
    numthrottles = 0;
    maxthrottles = 0;
    throttles = (throttletab*) 0;
    if ( throttlefile != (char*) 0 )
	read_throttlefile( throttlefile );

    /* Log file. */
    if ( logfile != (char*) 0 )
	{
	if ( strcmp( logfile, "/dev/null" ) == 0 )
	    {
	    no_log = 1;
	    logfp = (FILE*) 0;
	    }
	else
	    {
	    logfp = fopen( logfile, "a" );
	    if ( logfp == (FILE*) 0 )
		{
		syslog( LOG_CRIT, "%.80s - %m", logfile );
		perror( logfile );
		exit( 42 );
		}
	    (void) fcntl( fileno( logfp ), F_SETFD, 1 );
	    }
	}
    else
	logfp = (FILE*) 0;

    /* Figure out uid/gid from user. */
    pwd = getpwnam( user );
    if ( pwd == (struct passwd*) 0 )
	{
	syslog( LOG_CRIT, "unknown user - '%.80s'", user );
	(void) fprintf( stderr, "%s: unknown user - '%s'\n", argv0, user );
	exit( 43 );
	}
    uid = pwd->pw_uid;
    gid = pwd->pw_gid;

    /* Switch directories if requested. */
    if ( dir != (char*) 0 )
	{
	if ( chdir( dir ) < 0 )
	    {
	    syslog( LOG_CRIT, "chdir - %m" );
	    perror( "chdir" );
	    exit( 44 );
	    }
	}
#ifdef USE_USER_DIR
    else if ( getuid() == 0 )
	{
	/* No explicit directory was specified, we're root, and the
	** USE_USER_DIR option is set - switch to the specified user's
	** home dir.
	*/
	if ( chdir( pwd->pw_dir ) < 0 )
	    {
	    syslog( LOG_CRIT, "chdir - %m" );
	    perror( "chdir" );
	    exit( 45 );
	    }
	}
#endif /* USE_USER_DIR */

    /* Get current directory. */
    (void) getcwd( cwd, sizeof(cwd) - 1 );
    if ( cwd[strlen( cwd ) - 1] != '/' )
	(void) strcat( cwd, "/" );

    if ( ! debug )
	{
	/* We're not going to use stdin stdout or stderr from here on, so close
	** them to save file descriptors.
	*/
	(void) fclose( stdin );
	(void) fclose( stdout );
	(void) fclose( stderr );

	/* Daemonize - make ourselves a subprocess. */
#ifdef HAVE_DAEMON
	if ( daemon( 1, 1 ) < 0 )
	    {
	    syslog( LOG_CRIT, "daemon - %m" );
	    exit( 46 );
	    }
#else /* HAVE_DAEMON */
	switch ( fork() )
	    {
	    case 0:
	    break;
	    case -1:
	    syslog( LOG_CRIT, "fork - %m" );
	    exit( 47 );
	    default:
	    exit( 0 );
	    }
#ifdef HAVE_SETSID
        (void) setsid();
#endif /* HAVE_SETSID */
#endif /* HAVE_DAEMON */
	}
    else
	{
	/* Even if we don't daemonize, we still want to disown our parent
	** process.
	*/
#ifdef HAVE_SETSID
        (void) setsid();
#endif /* HAVE_SETSID */
	}

    if ( pidfile != (char*) 0 )
	{
	/* Write the PID file. */
	FILE* pidfp = fopen( pidfile, "w" );
	if ( pidfp == (FILE*) 0 )
	    {
	    syslog( LOG_CRIT, "%.80s - %m", pidfile );
	    exit( 48 );
	    }
	(void) fprintf( pidfp, "%d\n", (int) getpid() );
	(void) fclose( pidfp );
	}

    /* Chroot if requested. */
    if ( do_chroot )
	{
	if ( chroot( cwd ) < 0 )
	    {
	    syslog( LOG_CRIT, "chroot - %m" );
	    perror( "chroot" );
	    exit( 49 );
	    }
	(void) strcpy( cwd, "/" );
	/* Always chdir to / after a chroot. */
	if ( chdir( cwd ) < 0 )
	    {
	    syslog( LOG_CRIT, "chroot chdir - %m" );
	    perror( "chroot chdir" );
	    exit( 50 );
	    }
	}

    /* Set up to catch signals. */
#ifdef	SIGHUP
    (void) my_signal( SIGHUP,	handle_hup );
#ifdef notdef
    (void) my_signal( SIGHUP,	SIG_IGN );
#endif
#endif
#ifdef	SIGINT
    (void) my_signal( SIGINT,	handle_term );
#endif
#ifdef	SIGILL
    (void) my_signal( SIGILL,	handle_term );
#endif
#ifdef	SIGQUIT
    (void) my_signal( SIGQUIT,	handle_term );
#endif
#ifdef	SIGABRT
    (void) my_signal( SIGABRT,	handle_term );
#endif
#ifdef	SIGIOT
    (void) my_signal( SIGIOT,	handle_term );
#endif
#ifdef	SIGBUS
    (void) my_signal( SIGBUS,	handle_term );
#endif
#ifdef	SIGFPE
    (void) my_signal( SIGFPE,	handle_term );
#endif
#ifdef	SIGSEGV
    (void) my_signal( SIGSEGV,	handle_term );
#endif
#ifdef	SIGUSR1
    (void) my_signal( SIGUSR1,	handle_usr1 );
#endif
#ifdef	SIGUSR2
    (void) my_signal( SIGUSR2,	handle_usr2 );
#endif
#ifdef	SIGPIPE
    (void) my_signal( SIGPIPE,	SIG_IGN );	/* get EPIPE instead */
#endif
#ifdef	SIGALRM
    (void) my_signal( SIGALRM,	handle_term );
#endif
#ifdef	SIGTERM
    (void) my_signal( SIGTERM,	handle_term );
#endif
#ifdef	SIGTSTP
    (void) my_signal( SIGTSTP,	handle_term );
#endif
#ifdef	SIGTTIN
    (void) my_signal( SIGTTIN,	handle_term );
#endif
#ifdef	SIGTTOU
    (void) my_signal( SIGTTOU,	handle_term );
#endif
#ifdef	SIGURG
    (void) my_signal( SIGURG,	handle_term );
#endif
#ifdef	SIGXCPU
    (void) my_signal( SIGXCPU,	handle_term );
#endif
#ifdef	SIGXFSZ
    (void) my_signal( SIGXFSZ,	handle_term );
#endif
#ifdef	SIGVTALRM
    (void) my_signal( SIGVTALRM,handle_term );
#endif
#ifdef	SIGWINCH
    (void) my_signal( SIGWINCH,	SIG_IGN );
#endif
#ifdef	SIGPWR
    (void) my_signal( SIGPWR,	handle_term );
#endif
    // added by Jeff on 2003/8/14
    (void) my_signal( SIGUSR1,	handle_block_req );
    // end added
    /* zero flag to show start value */
    gotSigMask = 0;

    /* we enable close-on-exec if we can spawn CGI subprocesses */
    /* and fd-map is enabled */
    if ( cgi_pattern != (char*) 0  &&
	mmc_cfg_get_param( MMC_P_USE_SENDFILE ) )
	mmc_cfg_set_param( MMC_P_CLOSE_ON_EXEC, 1 );
	/* NOTE: no else, close-on-exec is disabled by default */

    /* log information */
    syslog( LOG_NOTICE,
    "DefExpireAge %d OccMmcTime %d",
	mmc_cfg_get_param( MMC_P_DEFAULT_EXPIRE_AGE ),
	OccasionalMmcTime
	);
    syslog( LOG_NOTICE,
    "DesMaxMallocBytes %d DesMaxMallocFiles %d DesMaxMappedFiles %d DesMaxOpenedFiles %d",
	mmc_cfg_get_param( MMC_P_DESIRED_MAX_MALLOC_BYTES ),
	mmc_cfg_get_param( MMC_P_DESIRED_MAX_MALLOC_FILES ),
	mmc_cfg_get_param( MMC_P_DESIRED_MAX_MAPPED_FILES ),
	mmc_cfg_get_param( MMC_P_DESIRED_MAX_OPENED_FILES )
	);
    syslog( LOG_NOTICE,
    "MaxFileSizeL0 %d MaxFileSizeL1 %d MaxFileSizeL2 %d MaxFileSizeL3 %d  %s (%s) %s",
	mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L0 ),
	mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L1 ),
	mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L2 ),
	mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L3 ),
	mmc_cfg_get_param( MMC_P_USE_SENDFILE ) > 0 ?
		"USE SENDFILE" :
		" NO sendfile",
	httpd_typeof_sendfile(),
	mmc_cfg_get_param( MMC_P_CLOSE_ON_EXEC ) > 0 ?
		"(close-on-exec)" :
		""
	);
    syslog( LOG_NOTICE,
    "IdleSendRespTmo %d IdleKeepAliveTmo %d",
	(int)(MlsIdleSendRespTimeLimit  / 1000L),
	(int)(MlsIdleKeepAliveTimeLimit / 1000L)
	);

    /* Initialize the timer package. */
    tmr_init();

    /* Initialize the HTTP layer.  Got to do this before giving up root,
    ** so that we can bind to a privileged port.
    */
    hs = httpd_initialize(
	hostname,
	(gotv4 ? &sa4 : (httpd_sockaddr*) 0),
	(gotv6 ? &sa6 : (httpd_sockaddr*) 0),
	port, cgi_pattern, charset, cwd, no_log, logfp, no_symlink, do_vhost,
	do_global_passwd, url_pattern, local_pattern, no_empty_referers );
    if ( hs == (httpd_server*) 0 )
	{
	exit( 51 );
	}
    /* Set up the occasional mmc timer. */
    if ( tmr_create( (struct timeval*) 0, occasional_mmc, JunkClientData,
		OccasionalMmcTime * 1000L, 1 ) == (Timer*) 0 )
	{
	syslog( LOG_CRIT, "tmr_create(occasional_mmc) failed" );
	exit( 52 );
	}

    /* Set up the occasional tmr timer. */
    if ( tmr_create( (struct timeval*) 0, occasional_tmr, JunkClientData,
		OCCASIONAL_TMR_TIME * 1000L, 1 ) == (Timer*) 0 )
	{
	syslog( LOG_CRIT, "tmr_create(occasional_tmr) failed" );
	exit( 52 );
	}

    if ( numthrottles > 0 )
	{
	/* Set up the throttles timer. */
	if ( tmr_create( (struct timeval*) 0, update_throttles, JunkClientData,
		THROTTLE_TIME * 1000L, 1 ) == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(update_throttles) failed" );
	    exit( 53 );
	    }
	}
#ifdef STATS_TIME
    /* Set up the stats timer. */
    if ( tmr_create( (struct timeval*) 0, show_stats, JunkClientData,
		STATS_TIME * 1000L, 1 ) == (Timer*) 0 )
	{
	syslog( LOG_CRIT, "tmr_create(show_stats) failed" );
	exit( 54 );
	}
#endif /* STATS_TIME */
    start_time = stats_time = time( (time_t*) 0 );
    stats_connections = stats_bytes = 0L;
    stats_simultaneous = 0;

    /* If we're root, try to become someone else. */
    if ( getuid() == 0 )
	{
	/* Set aux groups to null. */
	if ( setgroups( 0, (const gid_t*) 0 ) < 0 )
	    {
	    syslog( LOG_CRIT, "setgroups - %m" );
	    exit( 55 );
	    }
	/* Set primary group. */
	if ( setgid( gid ) < 0 )
	    {
	    syslog( LOG_CRIT, "setgid - %m" );
	    exit( 56 );
	    }
	/* Try setting aux groups correctly - not critical if this fails. */
	if ( initgroups( user, gid ) < 0 )
	    syslog( LOG_WARNING, "initgroups - %m" );
#ifdef HAVE_SETLOGIN
	/* Set login name. */
        (void) setlogin( user );
#endif /* HAVE_SETLOGIN */
	/* Set uid. */
	if ( setuid( uid ) < 0 )
	    {
	    syslog( LOG_CRIT, "setuid - %m" );
	    exit( 57 );
	    }
	/* Check for unnecessary security exposure. */
	if ( ! do_chroot )
	    syslog(
		LOG_WARNING,
		"started as root without requesting chroot(), warning only" );
	}

    /* Initialize our connections table. */
    maxconnects = fdwatch_get_nfiles();
    if ( maxconnects <= SPARE_FDS * 2)
	{ /* also maxconnects < 0 */
	syslog( LOG_CRIT, "fdwatch initialization failure: maxconnects %d",
		maxconnects );
	exit( 58 );
	}
    maxconnects -= SPARE_FDS;
    connects = NEW( connecttab, maxconnects );
    if ( connects == (connecttab*) 0 )
	{
	syslog( LOG_CRIT, "out of memory allocating a connecttab[%d]",
		maxconnects );
	exit( 59 );
	}
    syslog( LOG_NOTICE, "allocated connecttab[], maxconnects %d",
		maxconnects );
    memset( connects, 0, sizeof( *connects ) * maxconnects );
    for ( cnum = 0; cnum < maxconnects; ++cnum )
	{
	connects[cnum].conn_state = CNST_FREE;
	connects[cnum].hc = (httpd_conn*) 0;
	}
    numconnects = 0;
    httpd_conn_count = 0;

    if ( hs != (httpd_server*) 0 )
	{
	if ( hs->listen4_fd != -1 )
	    fdwatch_add_fd( hs->listen4_fd, (void*) 0, FDW_READ );
	if ( hs->listen6_fd != -1 )
	    fdwatch_add_fd( hs->listen6_fd, (void*) 0, FDW_READ );
	}

    {
        // get security from /etc/config/uLinux.conf
	extern int __security_level;
	char tmp[32]={0};
	Conf_Get_Field("/etc/config/uLinux.conf", "System", "Security Level", tmp, 31);
	__security_level=atoi(tmp);
    }
    /* Main loop. */
    (void) gettimeofday( &tv, (struct timezone*) 0 );
    while ( ( ! terminate ) || numconnects > 0 )
	{
	if( __block_request )
	{
	    usleep(300000);
	    continue;
	}
	if ( gotSigMask )
	{   /* we got one or more signal */

	    if ( gotSigMask & GOT_O_SIGHUP )
	    {
		gotSigMask &= ~GOT_O_SIGHUP;
		thttpd_log_reopen();
#ifdef DVR
		load_ip_security_rules(hs);
#endif
	    }

	    if ( gotSigMask & GOT_O_SIGUSR1 )
	    {
		gotSigMask &= ~GOT_O_SIGUSR1;
		if ( !terminate )
		{
		    terminate = 1;
		    if ( hs != (httpd_server*) 0 )
		    {
			httpd_terminate( hs );
			hs = (httpd_server*) 0;
		    }
		    /* log, so we know how many connections */
		    /* we have to wait for */
		    syslog( LOG_NOTICE,
			"got signal SIGUSR1 (numconnects %d)", numconnects);

		    /* if there are no connections we exit from while loop */

		    if ( numconnects == 0 )
			continue;
		}
	    }

	    if ( gotSigMask & GOT_O_SIGUSR2 )
	    {
		gotSigMask &= ~GOT_O_SIGUSR2;
		logstats( (struct timeval*) 0 );
	    }
	}
	/* Do the fd watch. */
	num_ready = fdwatch( tmr_mstimeout( &tv ) );
	if ( num_ready < 0 )
	    {
	    if ( errno == EINTR )
		continue;       /* try again */

	    syslog( LOG_ERR, "fdwatch - %m" );
	    exit( 60 );
	    }
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	//DebugOutput("sec: %ld, usec: %ld\n", tv.tv_sec, tv.tv_usec);

	if ( num_ready == 0 )
	    {
	    /* No fd's are ready - run the timers. */
	    //DebugOutput("tmr_run1\n");
	    tmr_run( &tv );
	    continue;
	    }

	/* Is it a new connection? */
	if ( hs != (httpd_server*) 0 && hs->listen6_fd != -1 &&
	     fdwatch_check_fd( hs->listen6_fd ) )
	    {
	    if ( handle_newconnect( &tv, hs->listen6_fd ) )
		/* Go around the loop and do another fdwatch, rather than
		** dropping through and processing existing connections.
		** New connections always get priority.
		*/
		continue;
	    }
	if ( hs != (httpd_server*) 0 && hs->listen4_fd != -1 &&
	     fdwatch_check_fd( hs->listen4_fd ) )
	    {
	    if ( handle_newconnect( &tv, hs->listen4_fd ) )
		/* Go around the loop and do another fdwatch, rather than
		** dropping through and processing existing connections.
		** New connections always get priority.
		*/
		continue;
	    }

	/* Find the connections that need servicing. */
	for ( ridx = 0; ridx < num_ready; ++ridx )
	    {
	    c = (connecttab*) fdwatch_get_client_data( ridx );
	    if ( c == (connecttab*) 0 )
		continue;
	    hc = c->hc;
	    if ( c->conn_state == CNST_READING &&
		 fdwatch_check_fd( hc->conn_fd ) )
		{
		handle_read( c, &tv );
#ifdef NAS
		hc->counter = 0;
#ifdef STATISTICS
		add_httpd_user(hc);
#endif
#endif
		}
	    else if ( c->conn_state == CNST_SENDING &&
		 fdwatch_check_fd( hc->conn_fd ) )
		{
		handle_send( c, &tv );
#ifdef NAS
		if ((++hc->counter) > 100) {
			hc->counter = 0;
#ifdef STATISTICS
			add_httpd_user(hc);
#endif
		}
#endif
		}
	    else if ( c->conn_state == CNST_SENDING_RESP &&
		 fdwatch_check_fd( hc->conn_fd ) )
		handle_send_resp( c, &tv );
	    else if ( c->conn_state == CNST_LINGERING &&
		 fdwatch_check_fd( hc->conn_fd ) )
		handle_linger( c, &tv );
	    }
	    //DebugOutput("tmr_run2\n");
	tmr_run( &tv );

	}

    /* The main loop terminated. */
    shut_down();
    syslog( LOG_NOTICE, "exiting" );
    closelog();
    exit( 0 + SIGUSR1 * terminate );
    }


static void
parse_args( int argc, char** argv )
    {
    int argn;

    debug = 0;
    port = DEFAULT_PORT;
    dir = (char*) 0;
#ifdef ALWAYS_CHROOT
    do_chroot = 1;
#else /* ALWAYS_CHROOT */
    do_chroot = 0;
#endif /* ALWAYS_CHROOT */
    no_log = 0;
    no_symlink = do_chroot;
#ifdef ALWAYS_VHOST
    do_vhost = 1;
#else /* ALWAYS_VHOST */
    do_vhost = 0;
#endif /* ALWAYS_VHOST */
#ifdef ALWAYS_GLOBAL_PASSWD
    do_global_passwd = 1;
#else /* ALWAYS_GLOBAL_PASSWD */
    do_global_passwd = 0;
#endif /* ALWAYS_GLOBAL_PASSWD */
#ifdef CGI_PATTERN
    cgi_pattern = CGI_PATTERN;
#else /* CGI_PATTERN */
    cgi_pattern = (char*) 0;
#endif /* CGI_PATTERN */
    url_pattern = (char*) 0;
    no_empty_referers = 0;
    local_pattern = (char*) 0;
    throttlefile = (char*) 0;
    hostname = (char*) 0;
    logfile = (char*) 0;
    pidfile = (char*) 0;
    user = DEFAULT_USER;
    charset = DEFAULT_CHARSET;
    /* set values (ex config.h) */
#ifdef	OCCASIONAL_MMC_TIME
    mmc_cfg_set_param( MMC_P_CLEANUP_TIME,
				OCCASIONAL_MMC_TIME );
#endif
#ifdef	DEFAULT_EXPIRE_AGE
    mmc_cfg_set_param( MMC_P_DEFAULT_EXPIRE_AGE,
				DEFAULT_EXPIRE_AGE );
#endif
#ifdef	DESIRED_MAX_MALLOC_BYTES
    mmc_cfg_set_param( MMC_P_DESIRED_MAX_MALLOC_BYTES,
				DESIRED_MAX_MALLOC_BYTES );
#endif
#ifdef	DESIRED_MAX_MALLOC_FILES
    mmc_cfg_set_param( MMC_P_DESIRED_MAX_MALLOC_FILES,
				DESIRED_MAX_MALLOC_FILES );
#endif
#ifdef	DESIRED_MAX_MAPPED_FILES
    mmc_cfg_set_param( MMC_P_DESIRED_MAX_MAPPED_FILES,
				DESIRED_MAX_MAPPED_FILES );
#endif
#ifdef	DESIRED_MAX_OPENED_FILES
    mmc_cfg_set_param( MMC_P_DESIRED_MAX_OPENED_FILES,
				DESIRED_MAX_OPENED_FILES );
#endif
#ifdef	USE_SENDFILE
    mmc_cfg_set_param( MMC_P_USE_SENDFILE,
				USE_SENDFILE );
#endif
#ifdef	MAX_FILE_SIZE_L0
    mmc_cfg_set_param( MMC_P_MAX_FILE_SIZE_L0,
				MAX_FILE_SIZE_L0 );
#endif
#ifdef	MAX_FILE_SIZE_L1
    mmc_cfg_set_param( MMC_P_MAX_FILE_SIZE_L1,
				MAX_FILE_SIZE_L1 );
#endif
#ifdef	MAX_FILE_SIZE_L2
    mmc_cfg_set_param( MMC_P_MAX_FILE_SIZE_L2,
				MAX_FILE_SIZE_L2 );
#endif
#ifdef	MAX_FILE_SIZE_L3
    mmc_cfg_set_param( MMC_P_MAX_FILE_SIZE_L3,
				MAX_FILE_SIZE_L3 );
#endif
    {
    int	DefaultExpireAge =
			mmc_cfg_get_param( MMC_P_DEFAULT_EXPIRE_AGE );
    int	DesiredMaxMallocBytes = /* value should always fit into a int */
			mmc_cfg_get_param( MMC_P_DESIRED_MAX_MALLOC_BYTES );
    int	DesiredMaxMallocFiles =
			mmc_cfg_get_param( MMC_P_DESIRED_MAX_MALLOC_FILES );
    int	DesiredMaxMappedFiles =
			mmc_cfg_get_param( MMC_P_DESIRED_MAX_MAPPED_FILES );
    int	DesiredMaxOpenedFiles =
			mmc_cfg_get_param( MMC_P_DESIRED_MAX_OPENED_FILES );

    if ( DefaultExpireAge < 1 ||
	 DesiredMaxMallocBytes < 1 ||
	 DesiredMaxMallocFiles < 1 ||
	 DesiredMaxMappedFiles < 1 ||
	 DesiredMaxOpenedFiles < 1 )
	{
	fprintf( stderr,
	"%s: bad mmc default parameter (%s %d %s %d %s %d %s %d %s %d)\n",
		argv0,
		"DefaultExpireAge",	DefaultExpireAge,
		"DesiredMaxMallocBytes",DesiredMaxMallocBytes,
		"DesiredMaxMallocFiles",DesiredMaxMallocFiles,
		"DesiredMaxMappedFiles",DesiredMaxMappedFiles,
		"DesiredMaxOpenedFiles",DesiredMaxOpenedFiles
		);
	exit( 60 );
	}
    }

    {
    int	MaxFileSizeL0 = mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L0 );
    int	MaxFileSizeL1 = mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L1 );
    int	MaxFileSizeL2 = mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L2 );
    int	MaxFileSizeL3 = mmc_cfg_get_param( MMC_P_MAX_FILE_SIZE_L3 );

    if ( MaxFileSizeL0 < 0 ||
	 MaxFileSizeL1 < 1 ||
	 MaxFileSizeL2 < 1 ||
	 MaxFileSizeL3 < 1 )
	{
	fprintf( stderr,
	"%s: bad mmc default parameter (%s %d %s %d %s %d %s %d)\n",
		argv0,
		"MaxFileSizeL0",	MaxFileSizeL0,
		"MaxFileSizeL1",	MaxFileSizeL1,
		"MaxFileSizeL2",	MaxFileSizeL2,
		"MaxFileSizeL3",	MaxFileSizeL3
		);
	exit( 60 );
	}
    }

    argn = 1;
    while ( argn < argc && argv[argn][0] == '-' )
	{
	if ( strcmp( argv[argn], "-C" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    read_config( argv[argn] );
	    }
	else if ( strcmp( argv[argn], "-p" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    port = atoi( argv[argn] );
	    }
	else if ( strcmp( argv[argn], "-d" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    dir = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-r" ) == 0 )
	    {
	    do_chroot = 1;
	    no_symlink = 1;
	    }
	else if ( strcmp( argv[argn], "-nor" ) == 0 )
	    {
	    do_chroot = 0;
	    no_symlink = 0;
	    }
	else if ( strcmp( argv[argn], "-s" ) == 0 )
	    no_symlink = 0;
	else if ( strcmp( argv[argn], "-nos" ) == 0 )
	    no_symlink = 1;
	else if ( strcmp( argv[argn], "-u" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    user = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-c" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    cgi_pattern = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-t" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    throttlefile = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-h" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    hostname = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-l" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    logfile = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-v" ) == 0 )
	    do_vhost = 1;
	else if ( strcmp( argv[argn], "-nov" ) == 0 )
	    do_vhost = 0;
	else if ( strcmp( argv[argn], "-g" ) == 0 )
	    do_global_passwd = 1;
	else if ( strcmp( argv[argn], "-nog" ) == 0 )
	    do_global_passwd = 0;
	else if ( strcmp( argv[argn], "-i" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    pidfile = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-T" ) == 0 && argn + 1 < argc )
	    {
	    ++argn;
	    charset = argv[argn];
	    }
	else if ( strcmp( argv[argn], "-V" ) == 0 )
	    {
	    (void) fprintf( stderr, "%s\n", SERVER_SOFTWARE );
	    exit( 0 );
	    }
	else if ( strcmp( argv[argn], "-D" ) == 0 )
	    debug = 1;
	else
	    usage();
	++argn;
	}
    if ( argn != argc )
	usage();
    }


static void
usage( void )
    {
    (void) fprintf( stderr,
"usage:  %s [-C configfile] [-p port] [-d dir] [-r|-nor] [-s|-nos] [-v|-nov] [-g|-nog] [-u user] [-c cgipat] [-t throttles] [-h host] [-l logfile] [-i pidfile] [-T charset] [-V] [-D]\n",
	argv0 );
    exit( 60 );
    }


#ifdef USE_LAYOUT
static void *
map_layoutfile( char* filename, int* pfilesize )
    {
    struct stat sb = { 0 };
    int   filefd = EOF;
    void *pfilebuf;

    /* sanity checks */
    if ( filename == NULL )
	return NULL;
    if ( filename[0] == '\0' )
	return NULL;
    if ( pfilesize == NULL )
	return NULL;

    /* get file size */
    if ( stat( filename, &sb ) != 0)
	return NULL;

    /* an empty file is not mmapped() */
    if ( sb.st_size == 0 )
	return NULL;

    if ( sb.st_size > 16384 )
	return NULL;

    /* alloc file contents (no fdmap) */
    if ( ( pfilebuf = malloc( sb.st_size + sizeof(long) ) ) == NULL )
	return NULL;

    if ( ( filefd = open( filename, O_RDONLY ) ) == -1 )
	{
	free( pfilebuf );
	return NULL;
	}

    if ( read( filefd, pfilebuf, sb.st_size ) != (ssize_t) sb.st_size )
	{
	free( pfilebuf );
	(void) close( filefd );
	return NULL;
	}

    (void) close( filefd );

    /* OK, return */

    *pfilesize = (int) sb.st_size;

    return pfilebuf;
    }
#endif /* USE_LAYOUT */


static void
read_config( char* filename )
    {
    FILE* fp;
    char line[8192];
    char* cp;
    char* cp2;
    char* name;
    char* value;

    fp = fopen( filename, "r" );
    if ( fp == (FILE*) 0 )
	{
	perror( filename );
	exit( 61 );
	}

    while ( fgets( line, sizeof(line), fp ) != (char*) 0 )
	{
	/* Trim comments. */
	if ( ( cp = strchr( line, '#' ) ) != (char*) 0 )
	    *cp = '\0';

	/* Split line into words. */
	for ( cp = line; *cp != '\0'; cp = cp2 )
	    {
	    /* Skip leading whitespace. */
	    cp += strspn( cp, " \t\n\r" );
	    /* Find next whitespace. */
	    cp2 = cp + strcspn( cp, " \t\n\r" );
	    /* Insert EOS and advance next-word pointer. */
	    while ( *cp2 == ' '  || *cp2 == '\t' ||
		    *cp2 == '\n' || *cp2 == '\r' )
		*cp2++ = '\0';
	    /* Split into name and value. */
	    name = cp;
	    value = strchr( name, '=' );
	    if ( value != (char*) 0 )
		*value++ = '\0';
	    /* Interpret. */
	    if ( strcasecmp( name, "debug" ) == 0 )
		{
		no_value_required( name, value );
		debug = 1;
		}
	    else if ( strcasecmp( name, "port" ) == 0 )
		{
		value_required( name, value );
		port = atoi( value );
		}
	    else if ( strcasecmp( name, "dir" ) == 0 )
		{
		value_required( name, value );
		dir = e_strdup( value );
		}
	    else if ( strcasecmp( name, "chroot" ) == 0 )
		{
		no_value_required( name, value );
		do_chroot = 1;
		no_symlink = 1;
		}
	    else if ( strcasecmp( name, "nochroot" ) == 0 )
		{
		no_value_required( name, value );
		do_chroot = 0;
		no_symlink = 0;
		}
	    else if ( strcasecmp( name, "symlink" ) == 0 )
		{
		no_value_required( name, value );
		no_symlink = 0;
		}
	    else if ( strcasecmp( name, "nosymlink" ) == 0 )
		{
		no_value_required( name, value );
		no_symlink = 1;
		}
	    else if ( strcasecmp( name, "symlinks" ) == 0 )
		{
		no_value_required( name, value );
		no_symlink = 0;
		}
	    else if ( strcasecmp( name, "nosymlinks" ) == 0 )
		{
		no_value_required( name, value );
		no_symlink = 1;
		}
	    else if ( strcasecmp( name, "user" ) == 0 )
		{
		value_required( name, value );
		user = e_strdup( value );
		}
	    else if ( strcasecmp( name, "cgipat" ) == 0 )
		{
		value_required( name, value );
		cgi_pattern = e_strdup( value );
		}
	    else if ( strcasecmp( name, "urlpat" ) == 0 )
		{
		value_required( name, value );
		url_pattern = e_strdup( value );
		}
	    else if ( strcasecmp( name, "noemptyreferers" ) == 0 )
		{
		no_value_required( name, value );
		no_empty_referers = 1;
		}
	    else if ( strcasecmp( name, "localpat" ) == 0 )
		{
		value_required( name, value );
		local_pattern = e_strdup( value );
		}
	    else if ( strcasecmp( name, "throttles" ) == 0 )
		{
		value_required( name, value );
		throttlefile = e_strdup( value );
		}
	    else if ( strcasecmp( name, "host" ) == 0 )
		{
		value_required( name, value );
		hostname = e_strdup( value );
		}
	    else if ( strcasecmp( name, "logfile" ) == 0 )
		{
		value_required( name, value );
		logfile = e_strdup( value );
		}
	    else if ( strcasecmp( name, "vhost" ) == 0 )
		{
		no_value_required( name, value );
		do_vhost = 1;
		}
	    else if ( strcasecmp( name, "novhost" ) == 0 )
		{
		no_value_required( name, value );
		do_vhost = 0;
		}
	    else if ( strcasecmp( name, "globalpasswd" ) == 0 )
		{
		no_value_required( name, value );
		do_global_passwd = 1;
		}
	    else if ( strcasecmp( name, "noglobalpasswd" ) == 0 )
		{
		no_value_required( name, value );
		do_global_passwd = 0;
		}
	    else if ( strcasecmp( name, "pidfile" ) == 0 )
		{
		value_required( name, value );
		pidfile = e_strdup( value );
		}
	    else if ( strcasecmp( name, "charset" ) == 0 )
		{
		value_required( name, value );
		charset = e_strdup( value );
		}
	    else if ( strcasecmp( name, "DEFAULT_EXPIRE_AGE" ) == 0 )
		{
		int	DefaultExpireAge;
		value_required( name, value );
		DefaultExpireAge = mmc_cfg_set_param(
				MMC_P_DEFAULT_EXPIRE_AGE,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "DESIRED_MAX_MALLOC_BYTES" ) == 0 )
		{
		int	DesiredMaxMallocBytes;
		value_required( name, value );
		DesiredMaxMallocBytes = mmc_cfg_set_param(
				MMC_P_DESIRED_MAX_MALLOC_BYTES,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "DESIRED_MAX_MALLOC_FILES" ) == 0 )
		{
		int	DesiredMaxMallocFiles;
		value_required( name, value );
		DesiredMaxMallocFiles = mmc_cfg_set_param(
				MMC_P_DESIRED_MAX_MALLOC_FILES,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "DESIRED_MAX_MAPPED_FILES" ) == 0 )
		{
		int	DesiredMaxMappedFiles;
		value_required( name, value );
		DesiredMaxMappedFiles = mmc_cfg_set_param(
				MMC_P_DESIRED_MAX_MAPPED_FILES,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "DESIRED_MAX_OPENED_FILES" ) == 0 )
		{
		int	DesiredMaxOpenedFiles;
		value_required( name, value );
		DesiredMaxOpenedFiles = mmc_cfg_set_param(
				MMC_P_DESIRED_MAX_OPENED_FILES,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "USE_SENDFILE" ) == 0 )
		{
		int	UseSendFile;
		UseSendFile = mmc_cfg_set_param(
				MMC_P_USE_SENDFILE,
				1 );
		}
	    else if ( strcasecmp( name, "MAX_FILE_SIZE_L0" ) == 0 )
		{
		int	MaxFileSizeL0;
		value_required( name, value );
		MaxFileSizeL0 = mmc_cfg_set_param(
				MMC_P_MAX_FILE_SIZE_L0,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "MAX_FILE_SIZE_L1" ) == 0 )
		{
		int	MaxFileSizeL1;
		value_required( name, value );
		MaxFileSizeL1 = mmc_cfg_set_param(
				MMC_P_MAX_FILE_SIZE_L1,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "MAX_FILE_SIZE_L2" ) == 0 )
		{
		int	MaxFileSizeL2;
		value_required( name, value );
		MaxFileSizeL2 = mmc_cfg_set_param(
				MMC_P_MAX_FILE_SIZE_L2,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "MAX_FILE_SIZE_L3" ) == 0 )
		{
		int	MaxFileSizeL3;
		value_required( name, value );
		MaxFileSizeL3 = mmc_cfg_set_param(
				MMC_P_MAX_FILE_SIZE_L3,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "OCCASIONAL_MMC_TIME" ) == 0 )
		{
		value_required( name, value );
		OccasionalMmcTime = mmc_cfg_set_param(
				MMC_P_CLEANUP_TIME,
				atoi( value ) );
		}
	    else if ( strcasecmp( name, "IDLE_SEND_RESP_TIMELIMIT" ) == 0 )
		{
		int	IdleSendRespTimeLimit;
		value_required( name, value );
		IdleSendRespTimeLimit = atoi( value );
		if ( IdleSendRespTimeLimit < 2 )
		     IdleSendRespTimeLimit = 2;
		else
		if ( IdleSendRespTimeLimit > 300 )
		     IdleSendRespTimeLimit = 300;
		MlsIdleSendRespTimeLimit = 1000L * IdleSendRespTimeLimit;
		}
	    else if ( strcasecmp( name, "IDLE_KEEPALIVE_TIMELIMIT" ) == 0 )
		{
		int	IdleKeepAliveTimeLimit;
		value_required( name, value );
		IdleKeepAliveTimeLimit = atoi( value );
		if ( IdleKeepAliveTimeLimit < 1 )
		     IdleKeepAliveTimeLimit = 1;
		else
		if ( IdleKeepAliveTimeLimit > 300 )
		     IdleKeepAliveTimeLimit = 300;
		MlsIdleKeepAliveTimeLimit = 1000L * IdleKeepAliveTimeLimit;
		}
#ifdef USE_LAYOUT
	    /* we assume to find out at most one file */
	    /* per type (header / footer) */
	    else if ( strcasecmp( name, "layout_header" ) == 0 )
		{
		value_required( name, value );
		/* avoid memory leak when there are multiple layout_header= */
		if ( lheaderfile != NULL )
		    free( lheaderfile );
		if ( lheaderfile_map != NULL )
		    free( lheaderfile_map );
		lheaderfile = value;
		lheaderfile_len = 0;
		if ( ( lheaderfile_map = (char*) map_layoutfile(
				lheaderfile, &lheaderfile_len ) ) == NULL )
		    lheaderfile = NULL;
		else
		    lheaderfile = e_strdup( value );
		}
	    else if ( strcasecmp( name, "layout_footer" ) == 0 )
		{
		value_required( name, value );
		/* avoid memory leak when there are multiple layout_footer= */
		if ( lfooterfile != NULL )
		    free( lfooterfile );
		if ( lfooterfile_map != NULL )
		    free( lfooterfile_map );
		lfooterfile = value;
		lfooterfile_len = 0;
		if ( ( lfooterfile_map = (char*) map_layoutfile(
				lfooterfile, &lfooterfile_len ) ) == NULL )
		    lfooterfile = NULL;
		else
		    lfooterfile = e_strdup( value );
		}
#endif /* USE_LAYOUT */
	    else
		{
		(void) fprintf(
		    stderr, "%s: unknown config option '%s'\n", argv0, name );
		exit( 62 );
		}
	    }
	}

    (void) fclose( fp );
    }


static void
value_required( char* name, char* value )
    {
    if ( value == (char*) 0 )
	{
	(void) fprintf(
	    stderr, "%s: value required for %s option\n", argv0, name );
	exit( 63 );
	}
    }


static void
no_value_required( char* name, char* value )
    {
    if ( value != (char*) 0 )
	{
	(void) fprintf(
	    stderr, "%s: no value required for %s option\n",
	    argv0, name );
	exit( 64 );
	}
    }


static char*
e_strdup( char* oldstr )
    {
    char* newstr;

    newstr = strdup( oldstr );
    if ( newstr == (char*) 0 )
	{
	syslog( LOG_CRIT, "out of memory copying a string" );
	(void) fprintf( stderr, "%s: out of memory copying a string\n", argv0 );
	exit( 65 );
	}
    return newstr;
    }


static void
lookup_hostname( httpd_sockaddr* sa4P, size_t sa4_len, int* gotv4P, httpd_sockaddr* sa6P, size_t sa6_len, int* gotv6P )
    {
#if defined(HAVE_GETADDRINFO) && defined(HAVE_GAI_STRERROR)

    struct addrinfo hints;
    struct addrinfo* ai;
    struct addrinfo* ai2;
    struct addrinfo* aiv4;
    struct addrinfo* aiv6;
    int gaierr;
    char strport[10];

    memset( &hints, 0, sizeof(hints) );
    hints.ai_family = AF_UNSPEC;
    hints.ai_flags = AI_PASSIVE;
    hints.ai_socktype = SOCK_STREAM;
    (void) snprintf( strport, sizeof(strport), "%d", port );
    if ( (gaierr = getaddrinfo( hostname, strport, &hints, &ai )) != 0 )
	{
	syslog(
	    LOG_CRIT, "getaddrinfo %.80s - %.80s",
	    hostname, gai_strerror( gaierr ) );
	(void) fprintf(
	    stderr, "%s: getaddrinfo %s - %s\n",
	    argv0, hostname, gai_strerror( gaierr ) );
	exit( 66 );
	}

    /* Find the first IPv4 and IPv6 entries. */
    aiv4 = (struct addrinfo*) 0;
    aiv6 = (struct addrinfo*) 0;
    for ( ai2 = ai; ai2 != (struct addrinfo*) 0; ai2 = ai2->ai_next )
	{
	switch ( ai2->ai_family )
	    {
	    case AF_INET:
	    if ( aiv4 == (struct addrinfo*) 0 )
		aiv4 = ai2;
	    break;
#if defined(AF_INET6) && defined(HAVE_SOCKADDR_IN6)
	    case AF_INET6:
	    if ( aiv6 == (struct addrinfo*) 0 )
		aiv6 = ai2;
	    break;
#endif /* AF_INET6 && HAVE_SOCKADDR_IN6 */
	    }
	}

    if ( aiv4 == (struct addrinfo*) 0 )
	*gotv4P = 0;
    else
	{
	if ( sa4_len < aiv4->ai_addrlen )
	    {
	    syslog(
		LOG_CRIT, "%.80s - sockaddr too small (%d < %d)",
		hostname, sa4_len, aiv4->ai_addrlen );
	    exit( 67 );
	    }
	memset( sa4P, 0, sa4_len );
	memcpy( sa4P, aiv4->ai_addr, aiv4->ai_addrlen );
	*gotv4P = 1;
	}
    if ( aiv6 == (struct addrinfo*) 0 )
	*gotv6P = 0;
    else
	{
	if ( sa6_len < aiv6->ai_addrlen )
	    {
	    syslog(
		LOG_CRIT, "%.80s - sockaddr too small (%d < %d)",
		hostname, sa6_len, aiv6->ai_addrlen );
	    exit( 68 );
	    }
	memset( sa6P, 0, sa6_len );
	memcpy( sa6P, aiv6->ai_addr, aiv6->ai_addrlen );
	*gotv6P = 1;
	}

    freeaddrinfo( ai );

#else /* HAVE_GETADDRINFO && HAVE_GAI_STRERROR */

    struct hostent* he;

    *gotv6P = 0;

    memset( sa4P, 0, sa4_len );
#ifdef notdef
    /* We don't really need to set sa_len. */
#ifdef HAVE_SA_LEN
    sa4P->sa_len = sa4_len;
#endif /* HAVE_SA_LEN */
#endif
    sa4P->sa.sa_family = AF_INET;
    if ( hostname == (char*) 0 )
	sa4P->sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
    else
	{
	sa4P->sa_in.sin_addr.s_addr = inet_addr( hostname );
	if ( (int) sa4P->sa_in.sin_addr.s_addr == -1 )
	    {
	    he = gethostbyname( hostname );
	    if ( he == (struct hostent*) 0 )
		{
#ifdef HAVE_HSTRERROR
		syslog(
		    LOG_CRIT, "gethostbyname %.80s - %.80s",
		    hostname, hstrerror( h_errno ) );
		(void) fprintf(
		    stderr, "%s: gethostbyname %s - %s\n",
		    argv0, hostname, hstrerror( h_errno ) );
#else /* HAVE_HSTRERROR */
		syslog( LOG_CRIT, "gethostbyname %.80s failed", hostname );
		(void) fprintf(
		    stderr, "%s: gethostbyname %s failed\n", argv0, hostname );
#endif /* HAVE_HSTRERROR */
		exit( 69 );
		}
	    if ( he->h_addrtype != AF_INET )
		{
		syslog( LOG_CRIT, "%.80s - non-IP network address", hostname );
		(void) fprintf(
		    stderr, "%s: %s - non-IP network address\n",
		    argv0, hostname );
		exit( 70 );
		}
	    (void) memcpy(
		&sa4P->sa_in.sin_addr.s_addr, he->h_addr, he->h_length );
	    }
	}
    sa4P->sa_in.sin_port = htons( port );
    *gotv4P = 1;

#endif /* HAVE_GETADDRINFO && HAVE_GAI_STRERROR */
    }


static void
read_throttlefile( char* throttlefile )
    {
    FILE* fp;
    char buf[5000];
    char* cp;
    int len;
    char pattern[5000];
    long limit;
    struct timeval tv;

    fp = fopen( throttlefile, "r" );
    if ( fp == (FILE*) 0 )
	{
	syslog( LOG_CRIT, "%.80s - %m", throttlefile );
	perror( throttlefile );
	exit( 71 );
	}

    (void) gettimeofday( &tv, (struct timezone*) 0 );

    while ( fgets( buf, sizeof(buf), fp ) != (char*) 0 )
	{
	/* Nuke comments. */
	cp = strchr( buf, '#' );
	if ( cp != (char*) 0 )
	    *cp = '\0';

	/* Nuke trailing whitespace. */
	len = strlen( buf );
	while ( len > 0 &&
		( buf[len-1] == ' ' || buf[len-1] == '\t' ||
		  buf[len-1] == '\n' || buf[len-1] == '\r' ) )
	    buf[--len] = '\0';

	/* Ignore empty lines. */
	if ( len == 0 )
	    continue;

	/* Parse line. */
	if ( sscanf( buf, " %4900[^ \t] %ld", pattern, &limit ) != 2 || limit <= 0 )
	    {
	    syslog( LOG_CRIT,
		"unparsable line in %.80s - %.80s", throttlefile, buf );
	    (void) fprintf( stderr,
		"%s: unparsable line in %.80s - %.80s\n",
		argv0, throttlefile, buf );
	    continue;
	    }

	/* Nuke any leading slashes in pattern. */
	if ( pattern[0] == '/' )
	    (void) strcpy( pattern, &pattern[1] );
	while ( ( cp = strstr( pattern, "|/" ) ) != (char*) 0 )
	    (void) strcpy( cp + 1, cp + 2 );

	/* Check for room in throttles. */
	if ( numthrottles >= maxthrottles )
	    {
	    if ( maxthrottles == 0 )
		{
		maxthrottles = 16;	/* arbitrary */
		throttles = NEW( throttletab, maxthrottles );
		}
	    else
		{
		maxthrottles *= 2;
		throttles = RENEW( throttles, throttletab, maxthrottles );
		}
	    if ( throttles == (throttletab*) 0 )
		{
		syslog( LOG_CRIT, "out of memory allocating a throttletab" );
		(void) fprintf(
		    stderr, "%s: out of memory allocating a throttletab\n",
		    argv0 );
		exit( 72 );
		}
	    }

	/* Add to table. */
	throttles[numthrottles].pattern = strdup( pattern );
	if ( throttles[numthrottles].pattern == (char*) 0 )
	    {
	    syslog( LOG_CRIT, "out of memory copying a throttle pattern" );
	    (void) fprintf(
		stderr, "%s: out of memory copying a throttle pattern\n",
		argv0 );
	    exit( 73 );
	    }
	throttles[numthrottles].limit = limit;
	throttles[numthrottles].rate = 0;
	throttles[numthrottles].bytes_since_avg = 0;
	throttles[numthrottles].num_sending = 0;

	++numthrottles;
	}
    (void) fclose( fp );
    }


static void
shut_down( void )
    {
    int cnum;
    struct timeval tv = { 0, 0 };

    in_shut_down = 1;

    (void) gettimeofday( &tv, (struct timezone*) 0 );
    logstats( &tv );

    if ( connects != (connecttab*) 0 )
	{
	for ( cnum = 0; cnum < maxconnects; ++cnum )
	    {
	    if ( connects[cnum].conn_state != CNST_FREE )
		{
		if ( connects[cnum].hc != (httpd_conn*) 0 )
		    {
		    httpd_complete_request( connects[cnum].hc, &tv, 1 );
		    httpd_close_conn( connects[cnum].hc, &tv );
		    }
		connects[cnum].conn_state = CNST_FREE;
		}

	    if ( connects[cnum].hc != (httpd_conn*) 0 )
		{
		httpd_destroy_conn( connects[cnum].hc );
		free( (void*) connects[cnum].hc );
		--httpd_conn_count;
		connects[cnum].hc = (httpd_conn*) 0;
		}
	    }

	if ( hs != (httpd_server*) 0 )
	    {
	    httpd_server* ths = hs;
	    hs = (httpd_server*) 0;
	    httpd_terminate( ths );
	    }

	mmc_destroy();
	tmr_destroy();

	free( (void*) connects );
	connects = (connecttab *) 0;
    }
    if ( throttles != (throttletab*) 0 )
	{
	free( (void*) throttles );
	throttles = (throttletab*) 0;
	}

    in_shut_down = 0;

    }


static int
handle_newconnect( struct timeval* tvP, int listen_fd )
    {
    int cnum;
    connecttab* c;
    ClientData client_data;

    /* This loops until the accept() fails, trying to start new
    ** connections as fast as possible so we don't overrun the
    ** listen queue.
    */
    for (;;)
	{
	/* Is there room in the connection table? */
	if ( numconnects >= maxconnects )
	    {
	    /* Out of connection slots.  Run the timers, then the
	    ** existing connections, and maybe we'll free up a slot
	    ** by the time we get back here.
	    **/
	    syslog( LOG_WARNING, "too many connections (%d) !",
		numconnects );
		//DebugOutput("tmr_run3\n");
	    tmr_run( tvP );
	    return 0;
	    }
	/* Find a free connection entry. */
	for ( cnum = 0; cnum < maxconnects; ++cnum )
	    if ( connects[cnum].conn_state == CNST_FREE )
		break;
	c = &connects[cnum];
	/* Make the httpd_conn if necessary. */
	if ( c->hc == (httpd_conn*) 0 )
	    {
	    c->hc = NEW( httpd_conn, 1 );
	    if ( c->hc == (httpd_conn*) 0 )
		{
		syslog( LOG_CRIT, "out of memory allocating an httpd_conn" );
		exit( 74 );
		}
	    c->hc->initialized = 0;
	    ++httpd_conn_count;
	    }

	//return 1;

	/* Get the connection. */
	switch ( httpd_get_conn( hs, listen_fd, c->hc ) )
	    {
	    case GC_FAIL:
	    case GC_NO_MORE:
	    return 1;
	    }
	c->conn_state = CNST_READING;
	++numconnects;

	c->numtnums = 0;
	c->keep_alive = 0;

	client_data.p = c;
	c->idle_read_timer = tmr_create(
	    tvP, idle_read_connection, client_data, IDLE_READ_TIMELIMIT * 1000L,
	    0 );
	if ( c->idle_read_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(idle_read_connection) failed" );
	    exit( 75 );
	    }
	c->idle_send_timer = (Timer*) 0;
	c->wakeup_timer = (Timer*) 0;
	c->linger_timer = (Timer*) 0;
	c->bytes_sent = 0;
#ifdef USE_LAYOUT
	c->hc->layout = 0;
	c->hc->lheaderfile_len = 0;
	c->hc->lfooterfile_len = 0;
#endif /* USE_LAYOUT */

	/* Set the connection file descriptor to no-delay mode. */
	httpd_set_ndelay( c->hc->conn_fd );

	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );

	++stats_connections;
	if ( numconnects > stats_simultaneous )
	    stats_simultaneous = numconnects;
	}
    }


static void
handle_buf_read( connecttab* c, struct timeval* tvP )
    {
    ClientData client_data;
    httpd_conn* hc = c->hc;
	int ret;
#ifdef DVR
	httpd_server* hs = hc->hs;
#endif


    /* Do we have a complete request yet? */
    switch ( httpd_got_request( hc ) )
	{
	case GR_NO_REQUEST:
	    return;
	case GR_BAD_REQUEST:
	    httpd_send_err( hc, 400, httpd_err400title, "",
			httpd_err400form, "" );
	    resp_clear_connection( c, tvP );
	    return;
	default:
	    /* OK, got request */
	    break;
	}

    /* Yes.  Try parsing and resolving it. */
    if ( httpd_parse_request( hc ) < 0 )
	{
	resp_clear_connection( c, tvP );
	return;
	}

#ifdef DVR
	//DebugOutput("%d\n", hs->rule_count);
	if (hs->rule_count > 0 && ((strstr(hc->encodedurl, ".cgi")) || (hc->send_jpeg_buffer == 1))) {
		int i, ok;
		unsigned long addr = ntohl(hc->client_addr.sa_in.sin_addr.s_addr);

		if (hs->rule_type == 1) {
			ok = 0;
			for (i = 0; i < hs->rule_count; i++) {
				switch (hs->prules[i].type) {
				case IPSEC_IP:
					if (hs->prules[i].ipaddr.s_addr == addr)
						ok = 1;
					break;
				case IPSEC_IP_RANGE:
					if ((hs->prules[i].ipaddr.s_addr & hs->prules[i].netmask.s_addr) == (addr & hs->prules[i].netmask.s_addr))
						ok = 1;
					break;
				}
				if (ok == 1)
					break;
			}
		}
		else {
			ok = 1;
			for (i = 0; i < hs->rule_count; i++) {
				switch (hs->prules[i].type) {
				case IPSEC_IP:
					if (hs->prules[i].ipaddr.s_addr == addr)
						ok = 0;
					break;
				case IPSEC_IP_RANGE:
					if ((hs->prules[i].ipaddr.s_addr & hs->prules[i].netmask.s_addr) == (addr & hs->prules[i].netmask.s_addr))
						ok = 0;
					break;
				}
				if (ok == 0)
					break;
			}
		}
		if (ok == 0) {
			httpd_send_err(hc, 403, "Forbidden", "", "", "");
			resp_clear_connection(c, tvP);
		   	return;
		}
	}
#endif

    /* Check the throttle table */
    if ( ! check_throttles( c ) )
	{
	httpd_send_err(
#if defined(NAS) || defined(DVR)
		hc, 503, httpd_err503title, "", httpd_err503form, "" );
#else
	    hc, 503, httpd_err503title, "", httpd_err503form, hc->encodedurl );
#endif
	resp_clear_connection( c, tvP );
	return;
	}

    /* Start the connection going. */
    if ( (ret = httpd_start_request( hc, tvP, numconnects, maxconnects )) < 0 )
	{
#if defined(NAS) || defined(DVR)
	// add by Kent Tsai
	// 999 is not enable the web filemanager
	if (ret == -999)
		httpd_send_err(hc, 999, httpd_err999title, "", httpd_err999form, "");
	else if (ret == -503)
		httpd_send_err(hc, 999, httpd_err503title, "", httpd_err503form, "");
	else if (ret == LOCAL_DISK_MONITOR_REDIRECT)
	{
		DebugOutput("send_redirect cgi-bin/cfg_sys_tool.cgi?Item=4 ...\n");
		send_redirect(hc, "/cgi-bin/cfg_sys_tool.cgi?Item=4");//"/cgi-bin/data/cmp/ch_info_fail.html");
		goto ok;
	}
	else if (ret == LOCAL_DISK_BUSY_REJECT)
	{
		DebugOutput("httpd_send_err %s ...\n", httpd_err998title);
		httpd_send_err(hc, STATUS_LOCAL_DISK_BUSY_REJECT, httpd_err998title, "", httpd_err998form, "");
	}
	else if (ret == LOCAL_DISK_MODULE_ERROR)
	{
		//DebugOutput("httpd_send_err %s ...\n", httpd_err997title);
		//httpd_send_err(hc, STATUS_LOCAL_DISK_MODULE_ERROR, httpd_err997title, "", httpd_err997form, "");
	}
	else if (ret == LOCAL_DISK_MASTER_CGI_REJECT)
	{
		DebugOutput("LOCAL_DISK_MASTER_CGI_REJECT\n");
		//fprintf(stdout, "%s", "\r\n");
	}

#endif
	/* Something went wrong.  Close down the connection. */
	resp_clear_connection( c, tvP );
	return;
	}

ok:
    /* Fill in bytes_to_send. */
#ifdef USE_LAYOUT
    if ( hc->layout && hc->got_range )
	{
	/* we assume that the following conditions have already been checked */
	/*    1) init_byte_loc   <= end_byte_loc */
	/*    2) lheaderfile_len <= end_byte_loc */
	/*    3) end_byte_loc    < lheaderfile_len + bytes_to_send */
	/*    4) bytes_to_send   > 0 (if bytes_to_send < 1 then no layout) */
	/* NOTE: send_mime() has already been called so we cannot change */
	/*       init_byte_loc and end_byte_loc here, they must be right ! */

	if ( hc->end_byte_loc < hc->lheaderfile_len + hc->bytes_to_send )
	    {
	    hc->lfooterfile_len = 0;
	    c->bytes_to_send = hc->end_byte_loc + 1 - hc->lheaderfile_len;
	    }
	else
	    c->bytes_to_send = hc->bytes_to_send;

	if ( hc->init_byte_loc < hc->lheaderfile_len )
	    /* c->bytes_sent == 0 */
	    hc->lheaderfile_len -= hc->init_byte_loc;
	else
	    { /* hc->lheaderfile_len = 0 is setted later */
	    if ( hc->init_byte_loc < hc->lheaderfile_len + hc->bytes_to_send )
		c->bytes_sent = hc->init_byte_loc - hc->lheaderfile_len;
	    else
		{
		c->bytes_sent = hc->bytes_to_send;
		hc->lfooterfile_len -= ( hc->init_byte_loc -
			( hc->lheaderfile_len + hc->bytes_to_send ) );
		}
	    hc->lheaderfile_len = 0;
	    }
	}
    else
#endif /* USE_LAYOUT */
    if ( hc->got_range )
	{
	c->bytes_sent = hc->init_byte_loc;
	c->bytes_to_send = hc->end_byte_loc + 1;
	}
    else
	c->bytes_to_send = hc->bytes_to_send;

    /* Check if it's already handled. */
    if ( hc->file_fd == EOF && hc->file_address == (char*) 0 )
	{
	/* No fd / file address means someone else is handling it. */
	c->bytes_sent = hc->bytes_sent;
	resp_clear_connection( c, tvP );
	return;
	}
#ifndef USE_LAYOUT
    if ( c->bytes_sent >= c->bytes_to_send )
#else  /* USE_LAYOUT */
    if ( c->bytes_sent >= c->bytes_to_send &&
	 hc->lheaderfile_len == 0 &&
	 hc->lfooterfile_len == 0 )
#endif /* USE_LAYOUT */
	{
	/* There's nothing to send. */
	resp_clear_connection( c, tvP );
	return;
	}

    /* Cool, we have a valid connection and a file to send to it. */
    c->conn_state = CNST_SENDING;
    c->started_at = tvP->tv_sec;
    c->wouldblock_delay = 0;
    client_data.p = c;
    tmr_cancel( c->idle_read_timer );
    c->idle_read_timer = (Timer*) 0;
    c->idle_send_timer = tmr_create(
	tvP, idle_send_connection, client_data, IDLE_SEND_TIMELIMIT * 1000L,
	0 );
    if ( c->idle_send_timer == (Timer*) 0 )
	{
	syslog( LOG_CRIT, "tmr_create(idle_send_connection) failed" );
	exit( 77 );
	}

    fdwatch_del_fd( hc->conn_fd );
    fdwatch_add_fd( hc->conn_fd, c, FDW_WRITE );

    /* don't call handle_send here ! */
    }


static void
handle_read( connecttab* c, struct timeval* tvP )
    {
    int sz;
    httpd_conn* hc = c->hc;

    /* Is there room in our buffer to read more bytes? */
    if ( hc->read_idx >= hc->read_size )
	{
	if ( hc->read_size >= 5632 )
	    {
	    httpd_send_err( hc, 400, httpd_err400title, "",
				httpd_err400form, "" );
	    resp_clear_connection( c, tvP );
	    return;
	    }
	httpd_realloc_str(
	    &hc->read_buf, &hc->read_size, hc->read_size + 1024 );
	}

    /* Read some more bytes. */
    sz = read(
	hc->conn_fd, &(hc->read_buf[hc->read_idx]),
	hc->read_size - hc->read_idx );
    if ( sz == 0 )
	{
	/* EOF */
	if ( !c->keep_alive )
	    /* complain about EOF on first partial request */
	    httpd_send_err(
		hc, 400, httpd_err400title, "", httpd_err400form, "" );
	resp_clear_connection( c, tvP );
	return;
	}
    if ( sz < 0 )
	{
	/* Ignore EWOULDBLOCK errors.  At first glance you would think that
	** connections returned by fdwatch as readable should never give an
	** EWOULDBLOCK; however, this apparently can happen if a packet gets
	** garbled.
	*/
	if ( errno != EWOULDBLOCK )
	    {
	    httpd_send_err(
		hc, 400, httpd_err400title, "", httpd_err400form, "" );
	    resp_clear_connection( c, tvP );
	    }
	/* else !EWOULDBLOCK, wait for another packet */
	return;
	}
    hc->read_idx += sz;

    /*
    ** if we start getting new data on this socket, "promote" it
    ** to a small read timeout.
    */
    if ( c->hc->keep_alive )
	{
	ClientData client_data;

	if ( c->idle_read_timer != (Timer*) 0 )
	    {
	    tmr_cancel( c->idle_read_timer );
	    c->idle_read_timer = (Timer*) 0;
	    }

	client_data.p = c;
	c->idle_read_timer = tmr_create(
	    tvP, idle_read_connection, client_data,
		( IDLE_READ_TIMELIMIT / 2 ) * 1000L, 0 );
	if ( c->idle_read_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(idle_read_connection) failed" );
	    exit( 76 );
	    }

    	c->hc->keep_alive = 0;
	}

    handle_buf_read( c, tvP );
    }


static void
handle_send( connecttab* c, struct timeval* tvP )
    {
    int sz, coast;
    ClientData client_data;
    time_t elapsed;
    httpd_conn* hc = c->hc;
#ifdef USE_LAYOUT
    int bytes_to_write = 0;
    int iovidx = 0;
    struct iovec iv[4];
#endif

    /* Are we done?  We don't check for headers because here */
    /* we always send non empty files. */
#ifndef USE_LAYOUT
    if ( c->bytes_sent >= c->bytes_to_send )
#else
    if ( c->bytes_sent >= c->bytes_to_send &&
	( hc->layout == 0 || hc->lfooterfile_len == 0 ) )
#endif /* USE_LAYOUT */
	{
	/* This reply is over and socket send buffer should be almost empty.
	** Now we can safely start reading next HTTP/1.1 request without
	** fears about timing out too early, see below similar check condition.
	*/
	clear_connection( c, tvP, DO_KEEP_ALIVE );
	return;
	}

#ifndef USE_LAYOUT

    /* Do we need to write the headers first? */
    if ( hc->responselen == 0 )
	{
	/* No, just write the file. */
	if ( hc->file_fd != EOF )
	    sz = httpd_sendfile( hc->conn_fd, hc->file_fd,
		(off_t) c->bytes_sent,
		MIN( c->bytes_to_send - c->bytes_sent, c->limit ) );
	else
	    sz = write(
		hc->conn_fd, &(hc->file_address[c->bytes_sent]),
		MIN( c->bytes_to_send - c->bytes_sent, c->limit ) );
	}
    else
	{
	/* Yes.  We'll combine headers and file into a single writev(),
	** hoping that this generates a single packet.
	*/
	if ( hc->file_fd != EOF )
	    {
	    /* send header (always in RAM) */
	    sz = write( hc->conn_fd, hc->response, hc->responselen );

	    if ( sz == hc->responselen &&
		 c->bytes_sent < c->bytes_to_send )
		{ /* we have just written all headers and there is */
		  /* something else to send, so try to send it */
		  /* before TCP/IP nagle timer expires (200 mlsec.) */
		int sz2;
		sz2 = httpd_sendfile( hc->conn_fd, hc->file_fd,
			(off_t) c->bytes_sent,
			MIN( c->bytes_to_send - c->bytes_sent, c->limit ) );
		if ( sz2 > 0 )
		    sz += sz2;
		/* otherwise ignore I/O errors */
		}
	    }
	else
	    {
	    struct iovec iv[2];

	    iv[0].iov_base = hc->response;
	    iv[0].iov_len = hc->responselen;
	    iv[1].iov_base = &(hc->file_address[c->bytes_sent]);
	    iv[1].iov_len = MIN( c->bytes_to_send - c->bytes_sent, c->limit );
	    sz = writev( hc->conn_fd, iv, 2 );
	    }
	}
#else
    /* USE_LAYOUT */

    /* we don't support sendfilev() (available in Solaris 8/9) */
    /* so we have to do some dirty "hack" when we use httpd_sendfile() */

    if ( hc->file_fd != EOF )
	{
	int hdrlen = 0;

	sz = 0;

	/* Do we have header bytes to send ? */
	if ( hc->responselen > 0 )
	    {
	    iv[iovidx].iov_base = hc->response;
	    iv[iovidx++].iov_len = hc->responselen;
	    hdrlen += hc->responselen;
	    }

	/* Do we have layout header bytes to send ? */
	if ( hc->layout && hc->lheaderfile_len > 0 )
	    {
	    iv[iovidx].iov_base = lheaderfile_map +
				( lheaderfile_len - hc->lheaderfile_len );
	    iv[iovidx++].iov_len = hc->lheaderfile_len;
	    hdrlen += hc->lheaderfile_len;
	    }

	/* if needed then write headers */
	if ( iovidx > 0 )
	    sz = writev( hc->conn_fd, iv, iovidx );

	/* if headers have been completely sent then send file content */
	if ( sz == hdrlen && c->bytes_sent < c->bytes_to_send )
	    {
	    int sz2;
	    bytes_to_write = MIN( c->bytes_to_send - c->bytes_sent, c->limit );
	    sz2 = httpd_sendfile( hc->conn_fd, hc->file_fd,
		(off_t) c->bytes_sent,
		bytes_to_write );
	    if ( hdrlen > 0 )
		{
		if ( sz2 > 0 )
		    /* OK, we successfully sent */
		    /* some file content after headers */
		    sz += sz2;
		else
		    /* otherwise ignore I/O errors */
		    bytes_to_write = 0;
		}
	    else
		/* headers were sent in previous calls */
		sz = sz2;
	    }
	else
	/* Do we have layout footer bytes to send ? */
	if ( c->bytes_sent >= c->bytes_to_send &&
	     hdrlen == 0 && hc->layout && hc->lfooterfile_len > 0)
	    {
	    sz = write( hc->conn_fd,
		lfooterfile_map + ( lfooterfile_len - hc->lfooterfile_len ),
		hc->lfooterfile_len );
	    }
	}
    else
	{
	/* Do we have header bytes to send ? */
	if ( hc->responselen > 0 )
	    {
	    iv[iovidx].iov_base = hc->response;
	    iv[iovidx++].iov_len = hc->responselen;
	    }

	/* Do we have layout header bytes to send ? */
	if ( hc->layout && hc->lheaderfile_len > 0 )
	    {
	    iv[iovidx].iov_base = lheaderfile_map +
				( lheaderfile_len - hc->lheaderfile_len );
	    iv[iovidx++].iov_len = hc->lheaderfile_len;
	    }

	/* Do we have file content bytes to send ? */
	if ( c->bytes_sent < c->bytes_to_send )
	    {
	    bytes_to_write = MIN( c->bytes_to_send - c->bytes_sent, c->limit );
	    iv[iovidx].iov_base = &(hc->file_address[c->bytes_sent]);
	    iv[iovidx++].iov_len = bytes_to_write;
	    }

	/* Do we have layout footer bytes to send ? */
	if ( hc->layout && hc->lfooterfile_len > 0)
	    {
	    iv[iovidx].iov_base = lfooterfile_map +
				( lfooterfile_len - hc->lfooterfile_len );
	    iv[iovidx++].iov_len = hc->lfooterfile_len;
	    }

	/* Do we need a single write ? */
	if ( iovidx == 1 )
	    /* Yes, then use write */
	    sz = write( hc->conn_fd, iv[0].iov_base, iv[0].iov_len );
	else
	    /* No.  We'll combine headers, layout and file into
	    ** a single writev(), hoping that this generates a single packet.
	    */
	    sz = writev( hc->conn_fd, iv, iovidx );
	}

#endif /* USE_LAYOUT */

    if ( sz == 0 ||
	 ( sz < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ) ) )
	{
	/* This shouldn't happen, but some kernels, e.g.
	** SunOS 4.1.x, are broken and select() says that
	** O_NDELAY sockets are always writable even when
	** they're actually not.
	**
	** Current workaround is to block sending on this
	** socket for a brief adaptively-tuned period.
	** Fortunately we already have all the necessary
	** blocking code, for use with throttling.
	*/
	c->wouldblock_delay += MIN_WOULDBLOCK_DELAY;
	c->conn_state = CNST_PAUSING;
	fdwatch_del_fd( hc->conn_fd );
	client_data.p = c;
	c->wakeup_timer = tmr_create(
	    tvP, wakeup_connection, client_data, c->wouldblock_delay, 0 );
	if ( c->wakeup_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(wakeup_connection) failed" );
	    exit( 78 );
	    }
	return;
	}
    if ( sz < 0 )
	{
	/* Something went wrong, close this connection.
	**
	** If it's just an EPIPE, don't bother logging, that
	** just means the client hung up on us.
	**
	** On some systems, write() occasionally gives an EINVAL.
	** Dunno why, something to do with the socket going
	** bad.  Anyway, we don't log those either.
	**
	** And ECONNRESET isn't interesting either.
	*/
	if ( errno != EPIPE && errno != EINVAL && errno != ECONNRESET )
	    syslog( LOG_ERR, "write - %m sending %.80s", hc->encodedurl );
	resp_clear_connection( c, tvP );
	return;
	}

    /* Ok, we wrote something. */
    tmr_reset( tvP, c->idle_send_timer );

    /* Was this a headers + file writev()? */
    if ( hc->responselen > 0 )
	{
	/* Yes; did we write only part of the headers? */
	if ( sz < hc->responselen )
	    {
	    /* Yes; move the unwritten part to the front of the buffer. */
	    int newlen = hc->responselen - sz;
	    (void) memmove( hc->response, &(hc->response[sz]), newlen );
	    hc->responselen = newlen;
	    sz = 0;
	    }
	else
	    {
	    /* Nope, we wrote the full headers, so adjust accordingly. */
	    sz -= hc->responselen;
	    hc->responselen = 0;
	    }
	}

#ifndef USE_LAYOUT
    /* And update how much of the file we wrote. */
    c->bytes_sent += sz;
    c->hc->bytes_sent += sz;
#else
    /* USE_LAYOUT */

    if ( hc->layout && hc->lheaderfile_len > 0 )
	/* was this a layout header write too ? */
	if ( sz < hc->lheaderfile_len )
	    {	/* did we finish writing it ? */
	    hc->lheaderfile_len -= sz;
	    sz = 0;
	    }
	else
	    {
	    sz -= hc->lheaderfile_len;
	    hc->lheaderfile_len = 0;
	    }

    if ( sz > 0 && c->bytes_sent < c->bytes_to_send )
	/* did we write file content ? */
	if ( sz < bytes_to_write )
	    {	/* update how much of the file we wrote. */
	    c->bytes_sent += sz;
	    c->hc->bytes_sent += sz;
	    sz = 0;
	    }
	else
	    {
	    sz -= bytes_to_write;
	    c->bytes_sent += bytes_to_write;
	    c->hc->bytes_sent += bytes_to_write;
	    }

    if ( sz > 0 && hc->layout && hc->lfooterfile_len > 0 )
	/* was this a layout footer write too ? */
	if ( sz < hc->lfooterfile_len )
	    {	/* did we finish writing it ? */
	    hc->lfooterfile_len -= sz;
	    sz = 0;
	    }
	else
	    {
	    sz -= hc->lfooterfile_len;
	    hc->lfooterfile_len = 0;
	    }
#endif /* USE_LAYOUT */

    /* Are we done?  We don't check for headers because */
    /* we always send non empty files. */
#ifndef USE_LAYOUT
    if ( c->bytes_sent >= c->bytes_to_send )
#else
    if ( c->bytes_sent >= c->bytes_to_send &&
	( hc->layout == 0 || hc->lfooterfile_len == 0 ) )
#endif /* USE_LAYOUT */
	{
	/* This reply is over, we clear connection here
	** if we don't do keep-alive or we do it and there is
	** at least a partial HTTP/1.1 pipelined request.
	*/
	if ( !c->hc->do_keep_alive ||
	    c->hc->checked_idx + 1 < c->hc->read_idx )
//	    c->hc->checked_idx + 2 < c->hc->read_idx )
	    clear_connection( c, tvP, DO_KEEP_ALIVE );
	/* else
	** we wait for next event of send socket buffer empty
	** and then we clear the request connection,
	** see above at the beginning of this function.
	*/
	return;
	}

    /* Tune the (blockheaded) wouldblock delay. */
    if ( c->wouldblock_delay > MIN_WOULDBLOCK_DELAY )
	c->wouldblock_delay -= MIN_WOULDBLOCK_DELAY;

    /* If we're throttling, check if we're sending too fast. */
    if ( c->limit != THROTTLE_NOLIMIT )
	{
	elapsed = tvP->tv_sec - c->started_at;
	if ( elapsed == 0 || c->hc->bytes_sent / elapsed > c->limit )
	    {
	    c->conn_state = CNST_PAUSING;
	    fdwatch_del_fd( hc->conn_fd );
	    /* When should we send the next c->limit bytes
	    ** to get back on schedule?  If less than a second
	    ** (integer math rounding), use 1/8 second.
	    */
	    coast = ( c->hc->bytes_sent + c->limit ) / c->limit - elapsed;
	    client_data.p = c;
	    c->wakeup_timer = tmr_create(
		tvP, wakeup_connection, client_data,
		coast > 0 ? ( coast * 1000L ) : 125L, 0 );
	    if ( c->wakeup_timer == (Timer*) 0 )
		{
		syslog( LOG_CRIT, "tmr_create(wakeup_connection) failed" );
		exit( 79 );
		}
	    }
	}
    }


static void
handle_send_resp( connecttab* c, struct timeval* tvP )
    {
    int sz;
    httpd_conn* hc = c->hc;

    /* Do we need to write the headers ? */
    if ( hc->responselen <= 0 )
	{
	/* No, just clear connection */
	clear_connection( c, tvP, NO_KEEP_ALIVE );
	return;
	}

    sz = write( hc->conn_fd, hc->response, hc->responselen );

    if ( sz == 0 ||
	 ( sz < 0 && ( errno == EWOULDBLOCK || errno == EAGAIN ) ) )
	{
	/* This shouldn't happen, but some kernels, e.g.
	** SunOS 4.1.x, are broken and select() says that
	** O_NDELAY sockets are always writable even when
	** they're actually not.
	**
	** We choose to not wait for sending a response.
	*/
	clear_connection( c, tvP, NO_KEEP_ALIVE );
	return;
	}

    if ( sz < 0 )
	{
	/* Something went wrong, close this connection.
	**
	** If it's just an EPIPE, don't bother logging, that
	** just means the client hung up on us.
	**
	** On some systems, write() occasionally gives an EINVAL.
	** Dunno why, something to do with the socket going
	** bad.  Anyway, we don't log those either.
	**
	** And ECONNRESET isn't interesting either.
	*/
	if ( errno != EPIPE && errno != EINVAL && errno != ECONNRESET )
	    syslog( LOG_ERR, "write - %m sending error response %.80s",
		hc->encodedurl );
	clear_connection( c, tvP, NO_KEEP_ALIVE );
	return;
	}

    /* Ok, we wrote something. */

    tmr_reset( tvP, c->idle_send_timer );

    if ( sz < hc->responselen )
    {
	/* Yes; move the unwritten part to the front of the buffer. */
	int newlen = hc->responselen - sz;
	(void) memmove( hc->response, &(hc->response[sz]), newlen );
	hc->responselen = newlen;
	sz = 0;
    }
    else
    {
	/* Nope, we wrote the full headers, so adjust accordingly. */
	sz -= hc->responselen;
	hc->responselen = 0;
	clear_connection( c, tvP, NO_KEEP_ALIVE );
	return;
    }

    }



static void
handle_linger( connecttab* c, struct timeval* tvP )
    {
    char buf[2048];
    int r;

    /* In lingering-close mode we just read and ignore bytes.  An error
    ** or EOF ends things, otherwise we go until a timeout.
    */
    r = read( c->hc->conn_fd, buf, sizeof(buf) );
    if ( r <= 0 )
	really_clear_connection( c, tvP );
    }


static int
check_throttles( connecttab* c )
    {
    int tnum;

    c->numtnums = 0;
    c->limit = THROTTLE_NOLIMIT;
    for ( tnum = 0; tnum < numthrottles && c->numtnums < MAXTHROTTLENUMS;
	  ++tnum )
	if ( match( throttles[tnum].pattern, c->hc->expnfilename ) )
	    {
	    /* If we're way over the limit, don't even start. */
	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
		return 0;
	    if ( throttles[tnum].num_sending < 0 )
		{
		syslog( LOG_ERR,
		"throttle sending count was negative - shouldn't happen!" );
		throttles[tnum].num_sending = 0;
		}
	    c->tnums[c->numtnums++] = tnum;
	    ++throttles[tnum].num_sending;
	    c->limit = MIN(
		c->limit, throttles[tnum].limit / throttles[tnum].num_sending );
	    }
    return 1;
    }


static void
clear_throttles( connecttab* c, struct timeval* tvP )
    {
    int i, tnum;

    for ( i = 0; i < c->numtnums; ++i )
	{
	tnum = c->tnums[i];
	--throttles[tnum].num_sending;
	throttles[tnum].bytes_since_avg += c->hc->bytes_sent;
	}
    }


static void
update_throttles( ClientData client_data, struct timeval* nowP )
    {
    int tnum;

    for ( tnum = 0; tnum < numthrottles; ++tnum )
	{
	throttles[tnum].rate =
	    ( 3 * throttles[tnum].rate +
	      throttles[tnum].bytes_since_avg / THROTTLE_TIME ) / 4;
	throttles[tnum].bytes_since_avg = 0;
	/* Log a warning message if necessary. */
	if ( throttles[tnum].rate > throttles[tnum].limit )
	    {
	    if ( throttles[tnum].rate > throttles[tnum].limit * 2 )
		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld GREATLY exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
	    else
		syslog( LOG_NOTICE, "throttle #%d '%.80s' rate %ld exceeding limit %ld", tnum, throttles[tnum].pattern, throttles[tnum].rate, throttles[tnum].limit );
	    }
	}
    }


static void
resp_clear_connection( connecttab* c, struct timeval* tvP )
    {
    httpd_conn* hc = c->hc;

    //DebugOutput("resp_clear_connection\n");

    if ( c->idle_read_timer != (Timer*) 0 )
	{
	tmr_cancel( c->idle_read_timer );
	c->idle_read_timer = (Timer*) 0;
	}
    if ( c->idle_send_timer != (Timer*) 0 )
	{
	tmr_cancel( c->idle_send_timer );
	c->idle_send_timer = (Timer*) 0;
	}
    if ( c->wakeup_timer != (Timer*) 0 )
	{
	tmr_cancel( c->wakeup_timer );
	c->wakeup_timer = (Timer*) 0;
	}
    if ( c->linger_timer != (Timer*) 0 )
	{
	tmr_cancel( c->linger_timer );
	c->linger_timer = (Timer*) 0;
	}

    /* If we haven't actually sent the buffered response yet, do so now.
    ** almost surely because of an error.
    ** NOTE: sending at most 1-2 KB of data should fit into default
    **		send buffer (4-16 KB), unless we are pipelining
    **		but in this case we don't care because we handle
    **		partial writes.
    */
    if ( hc->responselen > 0 )
	{
	ClientData client_data;

	c->conn_state = CNST_SENDING_RESP;

	client_data.p = c;
	c->idle_send_timer = tmr_create(
		tvP, idle_send_resp_connection, client_data,
		MlsIdleSendRespTimeLimit,
		0 );
	if ( c->idle_send_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT,
		"tmr_create(idle_send_resp_connection) failed" );
	    exit( 95 );
	    }
	/* reset watch events, just to be sure */

	fdwatch_del_fd( c->hc->conn_fd );
	fdwatch_add_fd( c->hc->conn_fd, c, FDW_WRITE );

	/* send error response */

	handle_send_resp ( c, tvP );
	return;
	}

    clear_connection( c, tvP, NO_KEEP_ALIVE );

    }


static void
clear_connection( connecttab* c, struct timeval* tvP, int doKeep )
    {
    ClientData client_data;

    	//DebugOutput("clear_connection\n");

    /* caller should have just managed to eventually send error response */
    /* so c->hc-> responselen should be == 0, */
    /* there is no need to use write here */

    if ( c->idle_read_timer != (Timer*) 0 )
	{
	tmr_cancel( c->idle_read_timer );
	c->idle_read_timer = (Timer*) 0;
	}
    if ( c->idle_send_timer != (Timer*) 0 )
	{
	tmr_cancel( c->idle_send_timer );
	c->idle_send_timer = (Timer*) 0;
	}
    if ( c->wakeup_timer != (Timer*) 0 )
	{
	tmr_cancel( c->wakeup_timer );
	c->wakeup_timer = (Timer*) 0;
	}
    if ( c->linger_timer != (Timer*) 0 )
	{
	tmr_cancel( c->linger_timer );
	c->linger_timer = (Timer*) 0;
	}

    /* This is our version of Apache's lingering_close() routine, which is
    ** their version of the often-broken SO_LINGER socket option.  For why
    ** this is necessary, see http://www.apache.org/docs/misc/fin_wait_2.html
    ** What we do is delay the actual closing for a few seconds, while reading
    ** any bytes that come over the connection.  However, we don't want to do
    ** this unless it's necessary, because it ties up a connection slot and
    ** file descriptor which means our maximum connection-handling rate
    ** is lower.  So, elsewhere we set a flag when we detect the few
    ** circumstances that make a lingering close necessary.  If the flag
    ** isn't set we do the real close now.
    */

    if ( c->hc->do_keep_alive && doKeep )
	{
	c->conn_state = CNST_READING;

	client_data.p = c;
	c->idle_read_timer = tmr_create(
	    tvP, idle_read_connection, client_data,
		MlsIdleKeepAliveTimeLimit,
	    0 );
	if ( c->idle_read_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(idle_read_connection) failed" );
	    exit( 80 );
	    }

	c->bytes_sent = 0;
	c->numtnums = 0;
	c->keep_alive = 1;

	httpd_complete_request( c->hc, tvP, 1 );

	fdwatch_del_fd( c->hc->conn_fd );
	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );

	/* we can call the following function because
	** httpd_got_request() certainly returned GR_GOT_REQUEST.
	*/
	httpd_request_reset2( c->hc );
    	c->hc->keep_alive = c->keep_alive;

	/* we might already have next request in buffer
	** if client is sending pipelined requests,
	** thus we make a call here;  if there is no data or not enough data
	** then we will wait till data arrives or read timer expires.
	** NOTE: be careful about recursive calls of same function.
	*/
	handle_buf_read( c, tvP );
	}
    else if ( c->hc->should_linger )
        {
	c->conn_state = CNST_LINGERING;

	client_data.p = c;
	c->linger_timer = tmr_create(
	    tvP, linger_clear_connection, client_data, LINGER_TIME * 1000L, 0 );
	if ( c->linger_timer == (Timer*) 0 )
	    {
	    syslog( LOG_CRIT, "tmr_create(linger_clear_connection) failed" );
	    exit( 81 );
	    }

    	httpd_complete_request( c->hc, tvP, 1 );

	fdwatch_del_fd( c->hc->conn_fd );
	fdwatch_add_fd( c->hc->conn_fd, c, FDW_READ );

	/* We don't send anymore data so we can half close socket */
	/* before lingering;  this should force some browsers (MSIE) */
	/* who ignore header "Connection: close", to close their connection */
	/* without waiting for LINGER_TIME seconds */

	httpd_close_conn_wr( c->hc );
	}
     else
        {
    	httpd_complete_request( c->hc, tvP, !c->keep_alive );
	really_clear_connection( c, tvP );
	}
    }


static void
really_clear_connection( connecttab* c, struct timeval* tvP )
    {
    	//DebugOutput("really_clear_connection\n");
    if ( c->conn_state == CNST_FREE )
	{   /* this should never happen */
	syslog( LOG_NOTICE,
	    "really_clear_connection, already FRED (numconnects %d)",
	    numconnects );
	return;
	}
    stats_bytes += c->bytes_to_send;
    fdwatch_del_fd( c->hc->conn_fd );
    httpd_close_conn( c->hc, tvP );
    clear_throttles( c, tvP );
    if ( c->linger_timer != (Timer*) 0 )
	{
	tmr_cancel( c->linger_timer );
	c->linger_timer = (Timer*) 0;
	}
    c->hc->should_linger = 0;
    c->conn_state = CNST_FREE;
    c->keep_alive = 0;
    --numconnects;
    }


static void
idle_read_connection( ClientData client_data, struct timeval* nowP )
    {
    connecttab* c;

    c = (connecttab*) client_data.p;
    c->idle_read_timer = (Timer*) 0;
    if ( c->conn_state != CNST_FREE )
	{
	if ( !c->keep_alive )
	    {
	    syslog( LOG_INFO,
		    "%.80s connection timed out reading",
		    httpd_ntoa( &c->hc->client_addr ) );
	    httpd_send_err( c->hc, 408, httpd_err408title, "",
				httpd_err408form, "" );
	    }
	resp_clear_connection( c, nowP );
	}
    }


static void
idle_send_connection( ClientData client_data, struct timeval* nowP )
    {
    connecttab* c;

    c = (connecttab*) client_data.p;
    c->idle_send_timer = (Timer*) 0;
    if ( c->conn_state != CNST_FREE )
	{
	syslog( LOG_INFO,
	    "%.80s connection timed out sending",
	    httpd_ntoa( &c->hc->client_addr ) );
	/* there is no need to send a response */
	clear_connection( c, nowP, NO_KEEP_ALIVE );
	}
    }


static void
idle_send_resp_connection( ClientData client_data, struct timeval* nowP )
    {
    connecttab* c;

    c = (connecttab*) client_data.p;
    c->idle_send_timer = (Timer*) 0;
    if ( c->conn_state != CNST_FREE )
	{
	httpd_conn* hc = c->hc;

	if ( c->conn_state == CNST_SENDING_RESP )
	    {
	    /* Do we need to write the headers ? */

	    if ( hc->responselen > 0 )
		{
		int	sz;

		/* yes, we need to, last chance to complete send ... */

		sz = write( hc->conn_fd, hc->response, hc->responselen );

		if ( sz == hc->responselen )
		    {	/* OK, we succeded, better late than never */
			hc->responselen = 0;
		    }
		else
		if ( sz < 0 )
		    {	/* no, error */
			syslog( LOG_ERR,
			"write - %m sending idle error response %.80s",
			hc->encodedurl );
		    }
		/* else partial write, well we tried ... */
		}
	    }

	if ( hc->responselen != 0 )
	    {
	    syslog( LOG_INFO,
	    "%.80s connection timed out sending error response (conn_state %d)",
	    httpd_ntoa( &c->hc->client_addr ), c->conn_state );
	    }
	clear_connection( c, nowP, NO_KEEP_ALIVE );
	}
    }


static void
wakeup_connection( ClientData client_data, struct timeval* nowP )
    {
    connecttab* c;

    c = (connecttab*) client_data.p;
    c->wakeup_timer = (Timer*) 0;
    if ( c->conn_state == CNST_PAUSING )
	{
	c->conn_state = CNST_SENDING;
	fdwatch_add_fd( c->hc->conn_fd, c, FDW_WRITE );
	}
    }

static void
linger_clear_connection( ClientData client_data, struct timeval* nowP )
    {
    connecttab* c;

    	//DebugOutput("linger_clear_connection\n");

    c = (connecttab*) client_data.p;
    c->linger_timer = (Timer*) 0;
    really_clear_connection( c, nowP );
    }


static void
occasional_mmc( ClientData client_data, struct timeval* nowP )
    {
    mmc_cleanup( nowP );
    }


static void
occasional_tmr( ClientData client_data, struct timeval* nowP )
    {
    tmr_cleanup();
    }


#ifdef STATS_TIME
static void
show_stats( ClientData client_data, struct timeval* nowP )
    {
    logstats( nowP );
    }
#endif /* STATS_TIME */


/* Generate debugging statistics syslog messages for all packages. */
static void
logstats( struct timeval* nowP )
    {
    struct timeval tv;
    time_t now;
    long up_secs, stats_secs;

    if ( nowP == (struct timeval*) 0 )
	{
	(void) gettimeofday( &tv, (struct timezone*) 0 );
	nowP = &tv;
	}
    now = nowP->tv_sec;
    up_secs = now - start_time;
    stats_secs = now - stats_time;
    if ( stats_secs == 0 )
	stats_secs = 1;	/* fudge */
    stats_time = now;
    syslog( LOG_NOTICE,
	"up %ld seconds, stats for %ld seconds:", up_secs, stats_secs );

    thttpd_logstats( stats_secs );
    httpd_logstats( stats_secs );
    mmc_logstats( stats_secs );
    fdwatch_logstats( stats_secs );
    tmr_logstats( stats_secs );
    }


/* Generate debugging statistics syslog message. */
static void
thttpd_logstats( long secs )
    {
    syslog( LOG_NOTICE,
	"  thttpd - %ld connections (%g/sec), %d max simultaneous, %ld bytes (%g/sec), %d httpd_conns allocated",
	stats_connections, (float) stats_connections / secs,
	stats_simultaneous, stats_bytes, (float) stats_bytes / secs,
	httpd_conn_count );
    stats_connections = stats_bytes = 0L;
    stats_simultaneous = 0;
    }
