/*
 * SCSI Media Changer device driver for Linux 2.4
 *
 *     (c) 1996-2002 Gerd Knorr <kraxel@bytesex.org>
 *
 */

#define VERSION "0.19"

#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/blk.h>
#include <linux/completion.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/chio.h>			/* here are all the ioctls */

#define MAJOR_NR	SCSI_CHANGER_MAJOR
#define DEVICE_NAME	"SCSI Changer"
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define CH_EXTRA_DEVS   2

#include "scsi.h"
#include "hosts.h"
#include "constants.h"
#include "ch.h"
#include <scsi/scsi_ioctl.h>

/* insmod options */
static int check_busy = 1;
static int debug = 0;
static int init = 1;
static int reinit = 1;

static int dt_id[8] = { -1, -1, -1, -1 };
static int dt_lun[8];

MODULE_SUPPORTED_DEVICE("sch");
MODULE_DESCRIPTION("device driver for scsi media changer devices");
MODULE_AUTHOR("Gerd Knorr <kraxel@bytesex.org>");
MODULE_LICENSE("GPL");

MODULE_PARM(check_busy,"i");
MODULE_PARM_DESC(check_busy, \
    "enable/disable busy check for data transfer elements (default: on)");

MODULE_PARM(init,"i");
MODULE_PARM_DESC(init, \
    "initialize element status on driver load (default: on)");

MODULE_PARM(reinit,"i");
MODULE_PARM_DESC(reinit, \
    "reinitialize element status on unit attention (default: on)");

MODULE_PARM(debug,"i");
MODULE_PARM_DESC(debug,"enable/disable debug messages, also prints more "
		 "detailed sense codes on scsi errors (default: off)");

MODULE_PARM(dt_id,"1-8i");
MODULE_PARM(dt_lun,"1-8i");

/* ------------------------------------------------------------------- */

#define MAX_RETRIES   1
#define CH_TIMEOUT    (3600 * HZ)

static int  ch_init(void);
static void ch_finish(void);
static int  ch_attach(Scsi_Device *);
static int  ch_detect(Scsi_Device *);
static void ch_detach(Scsi_Device *);
static int  ch_open(struct inode * inode, struct file * filp);
static int  ch_release(struct inode * inode, struct file * filp);
static int  ch_ioctl(struct inode * inode, struct file * filp,
		     unsigned int cmd, unsigned long arg);

struct Scsi_Device_Template ch_template =
{
	name:      "media changer",
	tag:       "sch",
	scsi_type: TYPE_MEDIUM_CHANGER,
	major:     SCSI_CHANGER_MAJOR,
	detect:    ch_detect,
	init:      ch_init,
	finish:    ch_finish,
	attach:    ch_attach,
	detach:    ch_detach
};

static struct file_operations changer_fops =
{
	owner:   THIS_MODULE,
	open:    ch_open,
	release: ch_release,
	ioctl:   ch_ioctl,
};

static struct {
	unsigned char  sense;
	unsigned char  asc;
	unsigned char  ascq;
	int		   errno;
} err[] = {
/* Just filled in what looks right. Hav'nt checked any standard paper for
   these errno assignments, so they may be wrong... */
{ ILLEGAL_REQUEST, 0x21,0x01/*Invalid element address*/,           EBADSLT },
{ ILLEGAL_REQUEST, 0x28,0x01/*Import or export element accessed*/, EBADE   },
{ ILLEGAL_REQUEST, 0x3B,0x0D/*Medium destination element full*/,   EXFULL  },
{ ILLEGAL_REQUEST, 0x3B,0x0E/*Medium source element empty*/,       EBADE   },

{ ILLEGAL_REQUEST, 0x20,0x00/*Invalid command operation code*/,    EBADRQC },
{ 0,0,0,0 }
};

Scsi_Changer * scsi_Changers = NULL;

/* ------------------------------------------------------------------- */

static void
ch_request_done (Scsi_Cmnd * SCpnt)
{
	struct request * req;
	
	req = &SCpnt->request;
	req->rq_status = RQ_SCSI_DONE; /* Busy, but indicate request done */
        if (req->waiting != NULL) {
		complete(req->waiting);
	}
}

static int
ch_do_cmd(int target, unsigned char * sr_cmd,
	  void * buffer, unsigned buflength)
{
	static u_char init_cmd[6] = { INITIALIZE_ELEMENT_STATUS };
	int errno, ua = 0, retries = 0;
	Scsi_Cmnd * SCpnt;
	
	if(!scsi_block_when_processing_errors(scsi_Changers[target].device))
		return -ENXIO;

	SCpnt = scsi_allocate_device(scsi_Changers[target].device, 1, TRUE);
 retry:
	errno = 0;
	{
		DECLARE_COMPLETION(wait);
		SCpnt->request.waiting = &wait;
		if (ua && reinit) {
			init_cmd[1] = scsi_Changers[target].device->lun << 5;
			scsi_do_cmd(SCpnt, (void *) init_cmd, NULL, 0,
				    ch_request_done, CH_TIMEOUT, MAX_RETRIES);
			wait_for_completion(&wait);
		}
		scsi_do_cmd(SCpnt, (void *) sr_cmd, buffer, buflength,
			    ch_request_done, CH_TIMEOUT, MAX_RETRIES);
		wait_for_completion(&wait);
		SCpnt->request.waiting = NULL;
	}
	
	if(driver_byte(SCpnt->result) != 0) {
		if (debug) {
			printk(DEVICE_NAME " [debug] error, command: ");
			print_command(sr_cmd);
			printk("\n");
			print_sense(DEVICE_NAME " [debug]", SCpnt);
		}
		/* Check to see if additional sense information is available */
		if (SCpnt->sense_buffer[7] > 5 &&
		    SCpnt->sense_buffer[12] != 0) {
			int i;
			for (i = 0; err[i].errno != 0; i++) {
				if (err[i].sense == SCpnt->sense_buffer[ 2] &&
				    err[i].asc   == SCpnt->sense_buffer[12] &&
				    err[i].ascq  == SCpnt->sense_buffer[13]) {
					errno = -err[i].errno;
					break;
				}
			}
		}
		if (errno == 0) {
			errno = -EIO;
			switch(SCpnt->sense_buffer[2] & 0xf) {
			case UNIT_ATTENTION:
				if (debug)
					printk(DEVICE_NAME ": got UNIT ATTENTION\n");
				scsi_Changers[target].unit_attention = 1;
				ua = 1;
				if (retries++ < 3)
					goto retry;
				break;
			case NOT_READY:
				printk(DEVICE_NAME " reports NOT READY.\n");
				break;
			case ILLEGAL_REQUEST:
				printk(DEVICE_NAME " reports ILLEGAL REQUEST.\n");
				break;
			default:
				printk(DEVICE_NAME " error, command: ");
				print_command(sr_cmd);
				printk("\n");
				print_sense(DEVICE_NAME, SCpnt);
			}
		}
	}
	
	scsi_release_command(SCpnt);
	return errno;
}

/* ------------------------------------------------------------------------ */

static Scsi_Device*
find_device(struct Scsi_Host *host, u_char channel, u_char id, u_char lun)
{
	Scsi_Device *ret;
	ret = host->host_queue;
	
	while (ret != NULL &&
	       (ret->host    != host || ret->channel != channel ||
		ret->id      != id   || ret->lun     != lun))
		ret = ret->next;
	
	return ret;
}

static int
ch_elem_to_typecode(int target, u_int elem)
{
	int i;
	
	for (i = 0; i < 4; i++) {
		if (elem >= scsi_Changers[target].firsts[i]  &&
		    elem <  scsi_Changers[target].firsts[i] +
	            scsi_Changers[target].counts[i])
			return i+1;
	}
	return 0;
}

static int
ch_read_element_status(int target, u_int elem, char *data)
{
	u_char  cmd[12];
	u_char  *buffer;
	int     result;
	
	buffer = (unsigned char *) scsi_malloc(512);
	if(!buffer) return -ENOMEM;
	
 retry:
	memset(cmd,0,sizeof(cmd));
	cmd[0] = READ_ELEMENT_STATUS;
	cmd[1] = (scsi_Changers[target].device->lun << 5) | 
		(scsi_Changers[target].voltags ? 0x10 : 0) |
		ch_elem_to_typecode(target,elem);
	cmd[2] = (elem >> 8) & 0xff;
	cmd[3] = elem        & 0xff;
	cmd[5] = 1;
	cmd[9] = 255;
	if (0 == (result = ch_do_cmd(target, cmd, buffer, 256))) {
		if (((buffer[16] << 8) | buffer[17]) != elem) {
			printk(KERN_DEBUG DEVICE_NAME
			       "Oops: asked for element 0x%02x, got 0x%02x\n",
			       elem,(buffer[16] << 8) | buffer[17]);
			scsi_free(buffer, 512);
			return -1;
		}
		memcpy(data,buffer+16,16);
	} else {
		if (scsi_Changers[target].voltags) {
			scsi_Changers[target].voltags = 0;
			printk(KERN_INFO DEVICE_NAME
			       ": device has no volume tag support\n");
			goto retry;
		}
		printk(KERN_WARNING DEVICE_NAME
		       ":  READ ELEMENT STATUS for element 0x%x failed\n",elem);
	}
	
	scsi_free(buffer, 512);
	return result;
}

static int 
ch_init_elem(int target)
{
	int err;
	u_char cmd[6];

	printk(KERN_INFO DEVICE_NAME "INITIALIZE ELEMENT STATUS, may take some time ...\n");
	memset(cmd,0,sizeof(cmd));
	cmd[0] = INITIALIZE_ELEMENT_STATUS;
	cmd[1] = scsi_Changers[target].device->lun << 5;
	err = ch_do_cmd(target, cmd, NULL, 0);
	printk(KERN_INFO DEVICE_NAME "... done\n");
	return err;
}

static int
ch_readconfig(int target)
{
	u_char  cmd[10], data[16];
	u_char  *buffer;
	int     result,i,id,lun;
	u_int   elem;
	
	buffer = (unsigned char *) scsi_malloc(512);
	memset(buffer,0,512);
	if(!buffer) return -ENOMEM;
	
	memset(cmd,0,sizeof(cmd));
	cmd[0] = MODE_SENSE;
	cmd[1] = scsi_Changers[target].device->lun << 5;
	cmd[2] = 0x1d;
	cmd[4] = 255;
	result = ch_do_cmd(target, cmd, buffer, 255);
	if (0 != result) {
		cmd[1] |= (1<<3);
		result  = ch_do_cmd(target, cmd, buffer, 255);
	}
	if (debug) {
		printk("element address assignment page: ");
		for (i = 0; i < buffer[0]+buffer[3]; i++)
			printk("%02x ",buffer[i]);
		printk("\n");
	}
	if (0 == result) {
		scsi_Changers[target].firsts[CHET_MT] =
			(buffer[buffer[3]+ 6] << 8) | buffer[buffer[3]+ 7];
		scsi_Changers[target].counts[CHET_MT] =
			(buffer[buffer[3]+ 8] << 8) | buffer[buffer[3]+ 9];
		scsi_Changers[target].firsts[CHET_ST] =
			(buffer[buffer[3]+10] << 8) | buffer[buffer[3]+11];
		scsi_Changers[target].counts[CHET_ST] =
			(buffer[buffer[3]+12] << 8) | buffer[buffer[3]+13];
		scsi_Changers[target].firsts[CHET_IE] =
			(buffer[buffer[3]+14] << 8) | buffer[buffer[3]+15];
		scsi_Changers[target].counts[CHET_IE] =
			(buffer[buffer[3]+16] << 8) | buffer[buffer[3]+17];
		scsi_Changers[target].firsts[CHET_DT] =
			(buffer[buffer[3]+18] << 8) | buffer[buffer[3]+19];
		scsi_Changers[target].counts[CHET_DT] =
			(buffer[buffer[3]+20] << 8) | buffer[buffer[3]+21];
		if (debug) {
			printk(DEVICE_NAME
			       ": mt:0x%x+%d st:0x%x+%d ie:0x%x+%d dt:0x%x+%d\n",
			       scsi_Changers[target].firsts[CHET_MT],
			       scsi_Changers[target].counts[CHET_MT],
			       scsi_Changers[target].firsts[CHET_ST],
			       scsi_Changers[target].counts[CHET_ST],
			       scsi_Changers[target].firsts[CHET_IE],
			       scsi_Changers[target].counts[CHET_IE],
			       scsi_Changers[target].firsts[CHET_DT],
			       scsi_Changers[target].counts[CHET_DT]);
		}
	} else {
		printk(KERN_WARNING DEVICE_NAME
		       ": reading element address assigment page failed!\n");
	}
	
	/* look up the devices of the data transfer elements */
	scsi_Changers[target].dt =
		kmalloc(scsi_Changers[target].counts[CHET_DT]*sizeof(Scsi_Device*),
			GFP_ATOMIC);
	for (elem = 0; elem < scsi_Changers[target].counts[CHET_DT]; elem++) {
		id  = -1;
		lun = 0;
		if (-1 != dt_id[elem]) {
			id  = dt_id[elem];
			lun = dt_lun[elem];
			printk(KERN_INFO DEVICE_NAME " data transfer element 0x%x: "
			       "[insmod option] ",
			       elem+scsi_Changers[target].firsts[CHET_DT]);
		} else if (0 != ch_read_element_status
			   (target,elem+scsi_Changers[target].firsts[CHET_DT],data)) {
			printk(DEVICE_NAME " Huh?: READ ELEMENT STATUS for "
			       "data transfer element 0x%x failed\n",
			       elem+scsi_Changers[target].firsts[CHET_DT]);
		} else {
			printk(KERN_INFO DEVICE_NAME " data transfer element 0x%x: ",
			       elem+scsi_Changers[target].firsts[CHET_DT]);
			if (data[6] & 0x80) {
				printk("not this SCSI bus\n");
				scsi_Changers[target].dt[elem] = NULL;
			} else if (0 == (data[6] & 0x30)) {
				printk("ID/LUN unknown\n");
				scsi_Changers[target].dt[elem] = NULL;
			} else {
				id  = scsi_Changers[target].device->id;
				lun = 0;
				if (data[6] & 0x20) id  = data[7];
				if (data[6] & 0x10) lun = data[6] & 7;
			}
		}
		if (-1 != id) {
			printk("ID %i, LUN %i, ",id,lun);
			scsi_Changers[target].dt[elem] =
				find_device(scsi_Changers[target].device->host,
					    scsi_Changers[target].device->channel,
					    id,lun);
			if (!scsi_Changers[target].dt[elem]) {
				/* should not happen */
				printk("Huh? device not found!\n");
			} else {
				printk("name: %8.8s %16.16s %4.4s\n",
				       scsi_Changers[target].dt[elem]->vendor,
				       scsi_Changers[target].dt[elem]->model,
				       scsi_Changers[target].dt[elem]->rev);
			}
		}
	}
	scsi_Changers[target].voltags = 1;
	scsi_free(buffer, 512);
	return 0;
}

/* ------------------------------------------------------------------------ */

static int
ch_position(int target, u_int trans, u_int elem, int rotate)
{
	u_char  cmd[10];
	
	if (debug)
		printk(KERN_DEBUG DEVICE_NAME " position: 0x%x\n",elem);
	if (0 == trans)
		trans = scsi_Changers[target].firsts[CHET_MT];
	memset(cmd,0,sizeof(cmd));
	cmd[0]  = POSITION_TO_ELEMENT;
	cmd[1]  = scsi_Changers[target].device->lun << 5;
	cmd[2]  = (trans >> 8) & 0xff;
	cmd[3]  =  trans       & 0xff;
	cmd[4]  = (elem  >> 8) & 0xff;
	cmd[5]  =  elem        & 0xff;
	cmd[8]  = rotate ? 1 : 0;
	return ch_do_cmd(target, cmd, NULL,0);
}

static int
ch_move(int target, u_int trans, u_int src, u_int dest, int rotate)
{
	u_char  cmd[12];
	
	if (debug)
		printk(KERN_DEBUG DEVICE_NAME " move: 0x%x => 0x%x\n",src,dest);
	if (0 == trans)
		trans = scsi_Changers[target].firsts[CHET_MT];
	memset(cmd,0,sizeof(cmd));
	cmd[0]  = MOVE_MEDIUM;
	cmd[1]  = scsi_Changers[target].device->lun << 5;
	cmd[2]  = (trans >> 8) & 0xff;
	cmd[3]  =  trans       & 0xff;
	cmd[4]  = (src   >> 8) & 0xff;
	cmd[5]  =  src         & 0xff;
	cmd[6]  = (dest  >> 8) & 0xff;
	cmd[7]  =  dest        & 0xff;
	cmd[10] = rotate ? 1 : 0;
	return ch_do_cmd(target, cmd, NULL,0);
}

static int
ch_exchange(int target, u_int trans, u_int src,
	    u_int dest1, u_int dest2, int rotate1, int rotate2)
{
	u_char  cmd[12];
	
	if (debug)
		printk(KERN_DEBUG DEVICE_NAME " exchange: 0x%x => 0x%x => 0x%x\n",
		       src,dest1,dest2);
	if (0 == trans)
		trans = scsi_Changers[target].firsts[CHET_MT];
	memset(cmd,0,sizeof(cmd));
	cmd[0]  = EXCHANGE_MEDIUM;
	cmd[1]  = scsi_Changers[target].device->lun << 5;
	cmd[2]  = (trans >> 8) & 0xff;
	cmd[3]  =  trans       & 0xff;
	cmd[4]  = (src   >> 8) & 0xff;
	cmd[5]  =  src         & 0xff;
	cmd[6]  = (dest1 >> 8) & 0xff;
	cmd[7]  =  dest1       & 0xff;
	cmd[8]  = (dest2 >> 8) & 0xff;
	cmd[9]  =  dest2       & 0xff;
	cmd[10] = (rotate1 ? 1 : 0) | (rotate2 ? 2 : 0);
	
	return ch_do_cmd(target, cmd, NULL,0);
}

/* ------------------------------------------------------------------------ */

static int
ch_release(struct inode * inode, struct file * filp)
{
	int minor = MINOR(inode->i_rdev);
	
	scsi_Changers[minor].device->access_count--;
	if (scsi_Changers[minor].device->host->hostt->module)
		__MOD_DEC_USE_COUNT(scsi_Changers[minor].device->host->hostt->module);
	return 0;
}

static int
ch_open(struct inode * inode, struct file * filp)
{    
	int minor = MINOR(inode->i_rdev);
	
	if(minor >= ch_template.nr_dev || !scsi_Changers[minor].device)
		return -ENXIO;
	
	scsi_Changers[minor].device->access_count++;
	if (scsi_Changers[minor].device->host->hostt->module)
		__MOD_INC_USE_COUNT(scsi_Changers[minor].device->host->hostt->module);
	
	return 0;
}

static int
ch_checkrange(int target, int type, int unit)
{
	if (type < 0 || type > 3 || unit < 0 ||
	    unit >= scsi_Changers[target].counts[type])
		return -1;
	return 0;
}

/* for data transfer elements: check if they are busy */
static int
ch_is_busy(int target, int type, int unit)
{
	if (!check_busy)
		return 0;
	if (type != CHET_DT)
		return 0;
	if (!scsi_Changers[target].dt[unit])
		return 0;
	return scsi_Changers[target].dt[unit]->access_count;
}

static int
ch_ioctl(struct inode * inode, struct file * filp,
	 unsigned int cmd, unsigned long arg)
{
	int target = MINOR(inode->i_rdev);
	
	if (target >= ch_template.nr_dev ||
	    !scsi_Changers[target].device) return -ENXIO;
	
	switch (cmd) {
	case CHIOGPARAMS:
	{
		struct changer_params params;
		
		params.cp_curpicker = 0;
		params.cp_npickers  = scsi_Changers[target].counts[CHET_MT];
		params.cp_nslots    = scsi_Changers[target].counts[CHET_ST];
		params.cp_nportals  = scsi_Changers[target].counts[CHET_IE];
		params.cp_ndrives   = scsi_Changers[target].counts[CHET_DT];
		
		if (copy_to_user((void *) arg, &params, sizeof(params)))
			return -EFAULT;
		return 0;
	}
	
	case CHIOPOSITION:
	{
		struct changer_position pos;
		
		if (copy_from_user(&pos, (void*)arg, sizeof (pos)))
			return -EFAULT;

		if (0 != ch_checkrange(target, pos.cp_type, pos.cp_unit)) {
			if (debug)
				printk(DEVICE_NAME ": CHIOPOSITION: invalid parameter\n");
			return -EBADSLT;
		}
		return ch_position(target,0,
				   scsi_Changers[target].firsts[pos.cp_type] + pos.cp_unit,
				   pos.cp_flags & CP_INVERT);
		return 0;
	}
	
	case CHIOMOVE:
	{
		struct changer_move mv;

		if (copy_from_user(&mv, (void*)arg, sizeof (mv)))
			return -EFAULT;

		if (0 != ch_checkrange(target, mv.cm_fromtype, mv.cm_fromunit) ||
		    0 != ch_checkrange(target, mv.cm_totype,   mv.cm_tounit  )) {
			if (debug)
				printk(DEVICE_NAME ": CHIOMOVE: invalid parameter\n");
			return -EBADSLT;
		}
		if (ch_is_busy(target, mv.cm_fromtype, mv.cm_fromunit) ||
		    ch_is_busy(target, mv.cm_totype,   mv.cm_tounit  ))
			return -EBUSY;
		
		return ch_move(target,0,
			       scsi_Changers[target].firsts[mv.cm_fromtype] + mv.cm_fromunit,
			       scsi_Changers[target].firsts[mv.cm_totype]   + mv.cm_tounit,
			       mv.cm_flags & CM_INVERT);
	}

#if 0
	case CHIOINIT:
		return ch_init_changer(target);
#endif

	case CHIOEXCHANGE:
	{
		struct changer_exchange mv;
		
		if (copy_from_user(&mv, (void*)arg, sizeof (mv)))
			return -EFAULT;

		if (0 != ch_checkrange(target, mv.ce_srctype,  mv.ce_srcunit ) ||
		    0 != ch_checkrange(target, mv.ce_fdsttype, mv.ce_fdstunit) ||
		    0 != ch_checkrange(target, mv.ce_sdsttype, mv.ce_sdstunit)) {
			if (debug)
				printk(DEVICE_NAME ": CHIOEXCHANGE: invalid parameter\n");
			return -EBADSLT;
		}
		if (0 != ch_is_busy(target, mv.ce_srctype,  mv.ce_srcunit ) ||
		    0 != ch_is_busy(target, mv.ce_fdsttype, mv.ce_fdstunit) ||
		    0 != ch_is_busy(target, mv.ce_sdsttype, mv.ce_sdstunit))
			return -EBUSY;
		
		return ch_exchange
			(target,0,
			 scsi_Changers[target].firsts[mv.ce_srctype]  + mv.ce_srcunit,
			 scsi_Changers[target].firsts[mv.ce_fdsttype] + mv.ce_fdstunit,
			 scsi_Changers[target].firsts[mv.ce_sdsttype] + mv.ce_sdstunit,
			 mv.ce_flags & CE_INVERT1, mv.ce_flags & CE_INVERT2);
	}

	case CHIOGSTATUS:
	{
		struct changer_element_status ces;
		u_char data[16];
		int i;
		
		if (copy_from_user(&ces, (void*)arg, sizeof (ces)))
			return -EFAULT;

		if (ces.ces_type < 0 || ces.ces_type > 3)
			return -EINVAL;
		
		for (i = 0; i < scsi_Changers[target].counts[ces.ces_type]; i++) {
			if (0 != ch_read_element_status
			    (target, scsi_Changers[target].firsts[ces.ces_type]+i,data))
				return -EIO;
			put_user(data[2], ces.ces_data+i);
			if (data[2] & CESTATUS_EXCEPT)
				printk(DEVICE_NAME ": element 0x%x: asc=0x%x, ascq=0x%x\n",
				       scsi_Changers[target].firsts[ces.ces_type]+i,
				       (int)data[4],(int)data[5]);
			if (0 != ch_read_element_status
			    (target, scsi_Changers[target].firsts[ces.ces_type]+i,data))
				return -EIO;
		}
		return 0;
	}
    
	case CHIOGELEM:
	{
		struct changer_get_element cge;
		u_char  cmd[12];
		u_char  *buffer;
		int     elem,result,i;
		
		if (copy_from_user(&cge, (void*)arg, sizeof (cge)))
			return -EFAULT;

		if (0 != ch_checkrange(target, cge.cge_type, cge.cge_unit))
			return -EINVAL;
		elem = scsi_Changers[target].firsts[cge.cge_type] + cge.cge_unit;
		
		buffer = (unsigned char *) scsi_malloc(512);
		if(!buffer) return -ENOMEM;
		
	voltag_retry:
		memset(cmd,0,sizeof(cmd));
		cmd[0] = READ_ELEMENT_STATUS;
		cmd[1] = (scsi_Changers[target].device->lun << 5) |
			(scsi_Changers[target].voltags ? 0x10 : 0) |
			ch_elem_to_typecode(target,elem);
		cmd[2] = (elem >> 8) & 0xff;
		cmd[3] = elem        & 0xff;
		cmd[5] = 1;
		cmd[9] = 255;
		
		if (0 == (result = ch_do_cmd(target, cmd, buffer, 256))) {
			cge.cge_status = buffer[18];
			cge.cge_flags = 0;
			if (buffer[18] & CESTATUS_EXCEPT) {
				/* FIXME: fill cge_errno */
			}
			if (buffer[25] & 0x80) {
				cge.cge_flags |= CGE_SRC;
				if (buffer[25] & 0x40)
					cge.cge_flags |= CGE_INVERT;
				elem = (buffer[26]<<8) | buffer[27];
				for (i = 0; i < 4; i++) {
					if (elem >= scsi_Changers[target].firsts[i] &&
					    elem < scsi_Changers[target].firsts[i]+
					    scsi_Changers[target].counts[i]) {
						cge.cge_srctype = i;
						cge.cge_srcunit = elem-scsi_Changers[target].firsts[i];
					}
				}
			}
			if ((buffer[22] & 0x30) == 0x30) {
				cge.cge_flags |= CGE_IDLUN;
				cge.cge_id  = buffer[23];
				cge.cge_lun = buffer[22] & 7;
			}
			if (buffer[9] & 0x80) {
				cge.cge_flags |= CGE_PVOLTAG;
				memcpy(cge.cge_pvoltag,buffer+28,36);
			}
			if (buffer[9] & 0x40) {
				cge.cge_flags |= CGE_AVOLTAG;
				memcpy(cge.cge_avoltag,buffer+64,36);
			}
		} else if (scsi_Changers[target].voltags) {
			scsi_Changers[target].voltags = 0;
			printk(KERN_INFO DEVICE_NAME
			       ": device has no volume tag support\n");
			goto voltag_retry;
		}
		scsi_free(buffer, 512);
		
		if (copy_to_user((void*)arg, &cge, sizeof (cge)))
			return -EFAULT;
		return result;
	}
    
	default:
		return scsi_ioctl(scsi_Changers[target].device, cmd, (void*)arg);

	}
}

/* ------------------------------------------------------------------------ */

static int
ch_detect(Scsi_Device * SDp)
{
	if(SDp->type != TYPE_MEDIUM_CHANGER)
		return 0;
	
	printk("Detected scsi changer ch%d at scsi%d, channel %d, id %d, lun %d\n", 
	       ch_template.dev_noticed++,
	       SDp->host->host_no, SDp->channel, SDp->id, SDp->lun); 
	
	return 1;
}

static int ch_attach(Scsi_Device * SDp){
	Scsi_Changer * cpnt;
	int i;
	
	if(SDp->type != TYPE_MEDIUM_CHANGER)
		return 1;
    
	if (ch_template.nr_dev >= ch_template.dev_max)
	{
		SDp->attached--;
		return 1;
	}
	
	for(cpnt = scsi_Changers, i=0; i<ch_template.dev_max; i++, cpnt++) 
		if(!cpnt->device) break;
	
	if (i >= ch_template.dev_max)
		panic ("scsi_devicelist corrupt (ch)");
	
	scsi_Changers[i].device = SDp;
	scsi_Changers[i].de =
		devfs_register (scsi_Changers[i].device->de, "changer",
				DEVFS_FL_DEFAULT, MAJOR_NR, i,
				S_IFCHR | S_IRUGO | S_IWUGO,
				&changer_fops, NULL);
	ch_template.nr_dev++;
	if(ch_template.nr_dev > ch_template.dev_max)
		panic ("scsi_devicelist corrupt (ch)");
	return 0;
}

static int ch_registered = 0;

static int ch_init()
{
	if (ch_template.dev_noticed == 0)
		return 0;

	if(!ch_registered) {
		if (devfs_register_chrdev(MAJOR_NR,"ch",&changer_fops)) {
			printk("Unable to get major %d for SCSI-Changer\n",MAJOR_NR);
			return 1;
		}
		ch_registered++;
	}
	
	if (scsi_Changers) return 0;
	ch_template.dev_max =
		ch_template.dev_noticed + CH_EXTRA_DEVS;
	scsi_Changers = (Scsi_Changer *)
		kmalloc(ch_template.dev_max * sizeof(Scsi_Changer), GFP_ATOMIC);
	memset(scsi_Changers, 0, ch_template.dev_max * sizeof(Scsi_Changer));
	
	return 0;
}

void
ch_finish()
{
	int i;
	
	for (i = 0; i < ch_template.nr_dev; ++i)
	{
		if (scsi_Changers[i].device) {
			ch_readconfig(i);
			if (init)
				ch_init_elem(i);
		}
	}
	
	return;
}	

static void ch_detach(Scsi_Device * SDp)
{
	Scsi_Changer * cpnt;
	int i;
	
	for(cpnt = scsi_Changers, i=0; i<ch_template.dev_max; i++, cpnt++) {
		if(cpnt->device == SDp) {
			/*
			 * Reset things back to a sane state so that one can re-load a new
			 * driver (perhaps the same one).
			 */
			devfs_unregister(cpnt->de);
			kfree(cpnt->dt);
			cpnt->device = NULL;
			SDp->attached--;
			ch_template.nr_dev--;
			ch_template.dev_noticed--;
			return;
		}
	}
	return;
}

static int __init init_ch_module(void)
{
	printk(KERN_INFO "SCSI Media Changer driver v" VERSION
	       " for Linux " UTS_RELEASE "\n");
	ch_template.module = THIS_MODULE;
	return scsi_register_module(MODULE_SCSI_DEV, &ch_template);
}

static void __exit exit_ch_module(void) 
{
	scsi_unregister_module(MODULE_SCSI_DEV, &ch_template);
	devfs_unregister_chrdev(MAJOR_NR, "ch");
	ch_registered--;
	if(scsi_Changers != NULL)
		kfree((char *) scsi_Changers);
	
	ch_template.dev_max = 0;
}

module_init(init_ch_module);
module_exit(exit_ch_module);
EXPORT_NO_SYMBOLS;

/*
 * Local variables:
 * c-basic-offset: 8
 * End:
 */
